/*
 * jtag.h : the head file for jtag.c
 * 
 * Copyright (C) 2003, Rongkai zhan <zhanrk@163.com>, All rights reserved.
 *
 * $Id: jtag.h,v 1.1.1.1 2004/09/26 03:59:34 zhanrk Exp $
 */

#ifndef JTAGER_JTAG_H
#define JTAGER_JTAG_H

#include "types.h"

/*
 * We use the Simple JTAG interface circuit, which is compatible with
 * Wiggler JTAG DONGLE, and its circuit schematic is referenced by the
 * file "doc/circuit.txt" or "doc/circuit.png".
 * 
 * Here is the SCHEMATIC for PARALLEL PORT <--> INTERFACE CIRCUIT
 * NOTE: Power comes from the JTAG connector (3.3 V) and feeds just
 * the 74HC240 and the pullups. The 74HC240 works OK at 3.3 Volts.
 * Instead of a 74HC240, a pair of 74HC14's can also be used. Put also
 * some decoupling on VDD (10 uF will do).

   JTAG VDD -------------------------------  +3.3V
   (PIN 1+2)          |                        |
                 74HC240 VDD                   |
                                              [5K]
                                               | 
   JTAG TMS  ________/|______/|_________[22K]__|___  D1 (LPT1)
    (PIN 7)          \|      \|                       (PIN 3)
                 74HC240    74HC240

                                             +3.3V
                                               |
                                              [5K] 
   JTAG TCK  ________/|______/|_________[22K]__|___  D2 (LPT1)
    (PIN 9)          \|      \|                       (PIN 4)
                 74HC240    74HC240
            
                                             +3.3V
                                               |
                                              [5K] 
   JTAG TDI  ________/|______/|_________[22K]__|___  D3 (LPT1)
    (PIN 5)          \|      \|                       (PIN 5)
                 74HC240    74HC240

   JTAG TDO  ________|\______|\__________________  BUSY (LPT1)
    (PIN 13)         |/      |/                       (PIN 11)
                 74HC240    74HC240
            
   JTAG GND  -------------------------------------- GND  (LPT1)
    (PIN 20)          |     |      |               (PINS 18..25)
                74HC240 E1  |  74HC240 E2
                            |
                        74HC240 GND
 * -------------------------------------------------------------
 */

/*
 * The bit map for the data port, status port of LPT1 (D-Type 25 pin connector),
 * which is defined by IEEE 1284 standard:
 *
 * data port (Out)		status port (In)
 * bit[7] -- pin9 (Out)		bit[7] -- pin11 (In), busy (Hardware Inverted)
 * bit[6] -- pin8 (Out)		bit[6] -- pin10 (In), Ack
 * bit[5] -- pin7 (Out)		bit[5] -- pin12 (In), Paper out
 * bit[4] -- pin6 (Out)		bit[4] -- pin13 (In), Select
 * bit[3] -- pin5 (Out)		bit[3] -- pin15 (In), Error
 * bit[2] -- pin4 (Out)		bit[2] -- IRQ(Not)
 * bit[1] -- pin3 (Out)		bit[1] -- Reserved
 * bit[0] -- pin2 (Out)		bit[0] -- Reserved
 *
 * NOTE: PIN11 is "hardware inverted", which means that the signal on PIN11
 * is inverted by the Parallel card's hardware. Such an example is the Busy
 * line. If +5v (Logic 1) was applied to this pin and the status register read,
 * it would return back a 0 in Bit 7 of the Status Register. In another words,
 * the bit 7 (busy) of status port is an active low input. E.g. If we read
 * a 0 from bit 7 of status port, this means that there is +5v at pin 11.
 */

/*
 * Masks definition for the parallel port pins
 * NOTE:TMS, TCK & TDI are connected to one pin each of D0-D7 (data port).
 * TDO is connected to one pin of the status port (BUSY)
 */
#define JTAG_TMS_MASK	0x02
#define JTAG_TCK_MASK	0x04
#define JTAG_TDI_MASK	0x08
#define JTAG_TRST_MASK	0x01 	/* optional TAP reset signal */

#define JTAG_TDO_MASK	0x80	/* remember this signal is inverted */

/* ARM7TDMI public jtag instructions */
#define JTAG_EXTEST 		0x00
#define JTAG_SCAN_N		0x02
#define JTAG_SAMPLE		0x03 /* SAMPLE/PRELOAD instruction */
#define JTAG_RESTART	0x04
#define JTAG_CLAMP		0x05
#define JTAG_HIGHZ		0x07
#define JTAG_CLAMPZ		0x09
#define JTAG_INTEST		0x0C
#define JTAG_IDCODE		0x0E
#define JTAG_BYPASS		0x0F	/* ·ָ*/

/*
 * The port I/O address of the Parallel port (LPTn), which we are using to
 * connect the JTAG dongle. By defaut, it is LPT1 whose I/O address is from
 * 0x378 to 0x380.
 */
extern unsigned int par_dataport;
extern unsigned int par_statusport;

extern int jtag_rw_dreg(int bitnr, const u32 *write_buf, u32 *read_buf);
extern int jtag_write_ireg(u32 jtag_instruction);

extern void enter_next_state(int tms);
extern int jtag_select_scanchain(u32 chain_num);
extern int jtag_idcode(u32 *idcode);
extern int jtag_bypass(void);
extern int jtag_reset(void);

#endif /* !JTAGER_JTAG_H */
