/*
 * target.h : the header file for the target definition
 * 
 * Copyright (C) 2003, Rongkai zhan <zhanrk@163.com>, All rights reserved.
 *
 * $Id: target.h,v 1.3 2004/10/17 13:54:32 zhanrk Exp $
 */
#ifndef JTAGER_TARGET_H
#define JTAGER_TARGET_H

#include "types.h"

/* core type definition */
#define TARGET_TYPE_ARM7TDMI	0
#define TARGET_TYPE_ARM9TDMI	1

/* core mode definition */
#define TARGET_MODE_ARM		0
#define TARGET_MODE_THUMB	1

/* target status definition */
#define TARGET_STATUS_RUNNING	0
#define TARGET_STATUS_HALTED	1

/* halt reason definition */
#define TARGET_HALTED_NONE	0 /* target is not halted */
#define TARGET_HALTED_BY_DBGRQ	1 /* target is halted by debug request */
#define TARGET_HALTED_BY_BREAKPT	2 /* target is halted by breakpoint */
#define TARGET_HALTED_BY_WATCHPT 	3 /* target is halted by watchpoint */

#define MAX_CORE_REGISTER_NUM 	20 /* the maximum number of core registers */
#define MAX_ICE_REGISTER_NUM    19 /* the maximum number of ice registers */

/* CPU core registers */
typedef struct core_register_struct {
	char *name;
	u32 value;
}core_register_t;

/* the register of the EmbeddedICE-RT logic */
typedef struct ice_register_struct {
	char *name;
	char *desc; /* long name string */
	u32 addr;
	int bitnr; /* the bit length of register */
	u32 value;
}ice_register_t;

/* the maxium words number of all scan chains test data.
 * ARM9TDMI scan chain 0 has 184 bits test data.
 */
#define MAX_SCANCHAIN_LENGTH		6

typedef struct scan_chain_struct {
	char *name; /* the name of scan chain */
	int bitnr; /* the bit number of testdata */

	/* the test data written in and the test data read out */
	u32 writein[MAX_SCANCHAIN_LENGTH];
	u32 readout[MAX_SCANCHAIN_LENGTH];

	/* points to the data struct specific to the scan chain */ 
	void *private;
}scan_chain_t;

/*
 * normal test data registers, such as: BYPASS register,
 * IDCODE register, INSTRUCTION register and SCAN PATH
 * SELECT register, etc.
 */
typedef struct testdata_register_struct {
	int bitnr; /* the bit length of test data register */
	u32 writein;
	u32 readout;
}testdata_reg_t;


#define MAX_SCANCHAIN_NUM  4

typedef struct target_struct {
	int type;	/* ARM7TDMI or ARM9TDMI */
	int mode;	/* ARM or Thumb */
	int status; /* halt, monitor or running */
	int halt_reason; /* BREAKPT/WATCHPT/DBGRQ */

	/* The core registers array and ICE registers array. 
	 * the last array element must be NULL.
	 */
	core_register_t *regs;
	ice_register_t  *ice_regs;
	
	/* target test data registers */
	testdata_reg_t bypass;	/* BYPASS regiseter */
	testdata_reg_t idcode;	/* Device ID code register */
	testdata_reg_t instruction;	/* instruction register */
	testdata_reg_t scanpath;	/* scan path select register */

	/* test data of all scan chains */
	u32 sc_num;	/* the number of scan chains of the target */
	u32 active_sc; /* the current active scan chain */
	scan_chain_t sc[MAX_SCANCHAIN_NUM];
	void *private;
}target_t;



/* the speed of instruction execution while in debug state */
#define DEBUG_SPEED	0
#define SYSTEM_SPEED	1

struct target_operation {
	/* ICE register read/write operations */
	int (*ice_read) (u32 reg_addr, u32 *reg_val);
	int (*ice_write) (u32 reg_addr, u32 reg_val);

	/* target cpu core operations */
	int (*halt) (void);
	int (*restart) (u32 pc);

	/* core registers read/write operations */
	int (*get_core_state) (void); /* read all core regs */
	int (*register_read) (core_register_t *reg);
	int (*register_write) (core_register_t *reg, u32 value);

	/* target memory read/write operations */
	int (*mem_read8) (u8 *buf, u32 address, u32 bytenr);
	int (*mem_write8) (u8 *buf, u32 address, u32 bytenr);
	int (*mem_read16) (u16 *buf, u32 address, u32 halfword_nr);
	int (*mem_write16) (u16 *buf, u32 address, u32 halfword_nr);
	int (*mem_read32) (u32 *buf, u32 address, u32 wordnr);
	int (*mem_write32) (u32 *buf, u32 address, u32 wordnr);
};

/* from target.c */
extern struct target_struct *target;
extern struct target_operation *t_op;

/* core register index */
#ifndef reg_index
#define reg_index(regname) (target_register_index((regname)))
#endif

/* ice register_index */
#ifndef ice_reg_index
#define ice_reg_index(regname) (target_ice_register_index((regname)))
#endif

extern core_register_t * target_register_index(const char *regname);
extern ice_register_t * target_ice_register_index(const char *regname);
extern int target_setup(int type); /* target initialization fuction */

/* ARM7TDMI core target */
#include "arm7tdmi.h"
#include "arm9tdmi.h"

/*
 * Memory I/O interface definition
 */
#undef read8
#undef read16
#undef read32
#undef write8
#undef write16
#undef write32

#define read8(__addr)   \
	({			\
		u8 __tmp = 0;	\
		t_op->mem_read8(&__tmp, __addr, 1); \
		__tmp; 		\
	})

#define read16(__addr)  \
	({			\
		u16 __tmp = 0;	\
		t_op->mem_read16(&__tmp, __addr, 1); \
		__tmp;		\
	})

#define read32(__addr)  \
	({			\
		u32 __tmp = 0;	\
		t_op->mem_read32(&__tmp, __addr, 1); \
		__tmp;		\
	})

#define write8(val, __addr) 	\
	({			\
		u8 __tmp = (u8)(val); \
		t_op->mem_write8(&__tmp, __addr, 1); \
	})

#define write16(val, __addr) 	\
	({			\
		u16 __tmp = (u16)(val); \
		t_op->mem_write16(&__tmp, __addr, 1); \
	})

#define write32(val, __addr) 	\
	({			\
		u32 __tmp = (u32)(val); \
		t_op->mem_write32(&__tmp, __addr, 1); \
	})

#endif /* !JTAGER_TARGET_H */
