/*
 * cmd/cmd_flerase.c : implements the flerase command
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_flerase.c,v 1.1.1.1 2004/09/26 03:59:47 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"
#include "target.h"
#include "flash.h"

#define FLERASE_SUBCMD_SECTOR  0x01
#define FLERASE_SUBCMD_CHIP    0x02

static int flerase_subcmd = 0;

static int whole_chip_flag = 0;         /* --whole-chip option */
static char *sector_address_str = NULL; /* --sector-address option */
static char *flerase_short_options = "a";
static struct option flerase_options[] = {
	{"whole-chip", no_argument, &whole_chip_flag, 1},
	{"sector-address", required_argument, 0, 'a'},
	{0, 0, 0, 0}
};

static struct option_desc flerase_optdesc[] = {
	{"whole-chip", "erase the whole FLASH chip."},
	{"sector-address", "the sector address of the FLASH sector to be erased."},
	{NULL, NULL}
};

cmd_t flerase_cmd;

static int do_flerase_sector(u32 sec_addr)
{
	int ret;

	if (sec_addr < flash->base) {
		printf("Error: address 0x%.8lX isn't within the address range"\
			" [0x%.8lX, 0x%.8lX) of current flash chip.\n",
			sec_addr, flash->base, flash->base+flash->chip_size);
		return -1;
	}

	/* sector address mask */
	sec_addr &= (~(u32)(flash->sector_size - 1));
	
	printf("Erasing %s flash sector [0x%.8lX, 0x%.8lX) ...... ",
		flash->name, sec_addr, (sec_addr + flash->sector_size));
	fflush(stdout);
	
	ret = flash->erase_sector(sec_addr);
	
	if (ret == 0)
		printf("[OK]\n");
	else
		printf("[FAILED]\n");
	return ret;
} /* end of do_flerase_sector(...) */

static int do_flerase_chip(void)
{
	int ret;

	ret = ask_yesno("Do you really want to erase the whole FLASH chip?");
	if (ret == 0)
		return ret; /* no */

	printf("Erase the whole flash chip %s: [0x%.8lX, 0x%.8lX) ...... ",
		flash->name, flash->base, flash->base+flash->chip_size);

	ret = flash->erase_chip();

	if (ret == 0)
		printf("[OK]\n");
	else
		printf("[FAILED]\n");
	return ret;
} /* end of do_flerase_sector(...) */

static int flerase_run(int argc, char **argv)
{
	int option_index = 0, ret = 0;
	u32 sector_addr;
	char *invalid_arg;
	
	if (argc == 1)
		return do_help(&flerase_cmd);
	
	/*
	 * Step 1: reset all global option flag variables
	 * and global option value strings
	 */
	whole_chip_flag = 0;
	sector_address_str = NULL;
	flerase_subcmd = 0;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, flerase_short_options,
				flerase_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (flerase_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 'a':
			/* --sector-address option */
			if (sector_address_str)
				goto syntax_err;
			sector_address_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (sector_address_str)
			goto syntax_err;
		sector_address_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3 - Translate command parameters
	 */
	if (whole_chip_flag) {
		if (argc > 2)
			goto syntax_err;
		flerase_subcmd = FLERASE_SUBCMD_CHIP;
	} else {
		if (sector_address_str == NULL)
			goto syntax_err;
		
		invalid_arg = sector_address_str;
		ret = str2int(sector_address_str, &sector_addr);
		if (ret)
			goto arg_err;
		flerase_subcmd = FLERASE_SUBCMD_SECTOR;
	}
	
	if (flash == NULL)
		goto no_flash;
	
	if (target->status == TARGET_STATUS_RUNNING)
		goto target_err;
	
	/*
	 * Step 4: go, go, go ...
	 */
	switch (flerase_subcmd) {
	case FLERASE_SUBCMD_SECTOR:
		ret = do_flerase_sector(sector_addr);
		break;
	case FLERASE_SUBCMD_CHIP:
		ret = do_flerase_chip();
		break;
	default:
		goto syntax_err;
	} /* end of switch(...) */

	return ret;
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
no_flash:
	printf("Error: no FLASH chip is selected currently. Please use "\
		"\"flprobe\" first.\n");
	return -1;
target_err:
	printf("error: %s\n", err_str(-ERR_TARGET_IS_RUNNING));
	return -1;
} /* end of flerase_run() */

cmd_t flerase_cmd = {
	.name = "flerase",
	.desc = "Erase one FLASH sector or the whole chip.",
	.alias = NULL,
	.usage = "flerase --whole-chip\n\tflerase [--sector-address=]addr",
	.verbose = NULL,
	.long_options = flerase_options,
	.optdesc = flerase_optdesc,
	.non_option_args = 1,

	.run = flerase_run,
};
