/*
 * cmd/cmd_flprobe.c : implements the flprobe command
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_flprobe.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"
#include "target.h"
#include "flash.h"

static char *chip_base_str = NULL;  /* --chip-base option */
static char *flprobe_short_options = "b";
static struct option flprobe_options[] = {
	{"chip-base", required_argument, 0, 'b'},
	{0, 0, 0, 0}
};

static struct option_desc flprobe_optdesc[] = {
	{"chip-base", "The base address of the whole flash chip's memory range."},
	{NULL, NULL}
};

cmd_t flprobe_cmd;

static int do_flash_probe(u32 base)
{
	int i, ret;
	flash_t *f;

	i = 0;
	flash = NULL;
	
	printf("Probing flash chip at 0x%.8lX ... ", base);
	fflush(stdout);
	while ((f = flashes[i])) {
		ret = f->probe(base);
		if (ret == 0) {
			flash = f;
			break;
		}
		i++;
	}

	if (ret) {
		printf("[NONE]\n");
		return ret;
	}
	printf("[FOUND : %s]\n", flash->name);
	printf("base address: 0x%.8lX\n", flash->base);
	printf("   chip size: %u KB\n", (flash->chip_size/1024));
	printf(" sector size: %u KB\n", (flash->sector_size/1024));
	printf("   bus width: %d-bit\n", flash->bus_width);
	return ret;
} /* end of do_flash_probe(...) */

static int flprobe_run(int argc, char **argv)
{
	int option_index = 0, ret = 0;
	u32 base;
	char *invalid_arg;
	
	/*
	 * Step 1: reset all global option flag variables
	 * and global option value strings
	 */
	chip_base_str = NULL;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, flprobe_short_options,
				flprobe_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (flprobe_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 'b':
			/* --base option */
			if (chip_base_str)
				goto syntax_err;
			chip_base_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (chip_base_str)
			goto syntax_err;
		chip_base_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3 - Translate command parameters
	 */
	base = 0x0; /* default base address: 0x00000000 */
	if (chip_base_str) {
		invalid_arg = chip_base_str;
		ret = str2int(chip_base_str, &base);
		if (ret)
			goto arg_err;
	}
	
	if (target->status == TARGET_STATUS_RUNNING)
		goto target_err;
	
	/*
	 * Step 4: go, go, go ...
	 */
	return do_flash_probe(base);
	
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
target_err:
	printf("error: %s\n", err_str(-ERR_TARGET_IS_RUNNING));
	return -1;
} /* end of flprobe_run() */

cmd_t flprobe_cmd = {
	.name = "flprobe",
	.desc = "Probe flash chips at the specified base address.",
	.alias = NULL,
	.usage = "flprobe [[--chip-base=]addr]",
	.verbose = NULL,
	.long_options = flprobe_options,
	.optdesc = flprobe_optdesc,
	.non_option_args = 1,

	.run = flprobe_run,
};
