/*
 * cmd/cmd_flwrite.c : implements the flwrite command
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_flwrite.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "jtager.h"
#include "target.h"
#include "flash.h"

#define FLWRITE_SUBCMD_MEMORY  0x01  /* The source data is from a target memory block */
#define FLWRITE_SUBCMD_INFILE  0x02  /* The source data is from a host binary file */

static int flwrite_subcmd = 0;

static char *to_str = NULL;   /* --to option */
static char *from_str = NULL; /* --from option */
static char *size_str = NULL; /* --size option */
static char *infile_str = NULL; /* --infile option */
static char *flwrite_short_options = "tfsi";
static struct option flwrite_options[] = {
	{"to", required_argument, 0, 't'},
	{"from", required_argument, 0, 'f'},
	{"size", required_argument, 0, 's'},
	{"infile", required_argument, 0, 'i'},
	{0, 0, 0, 0}
};

static struct option_desc flwrite_optdesc[] = {
	{"to", "The destination address within the FLASH address space."},
	{"from", "The base address of the source memory block."},
	{"size", "The size of the memory block, default: 1 word(32-bit)."},
	{"infile", "The pathname of the input data file"},
	{NULL, NULL}
};

cmd_t flwrite_cmd;

#define SECTOR_MASK (~(flash->sector_size - 1))

/*
 * do_flwrite_infile - write flash from host data file
 */
static int do_flwrite_infile(u32 dest, char *infile)
{
	struct stat infile_stat;
	int fd, secnr, i, ret;
	unsigned long total;
	unsigned long chip_start, chip_end;
	unsigned long start, end;
	ssize_t buf_len;
	u8 buf[PAGE_SIZE];
	
	ret = stat(infile, &infile_stat);
	if (ret)
		goto infile_err;
	fd = open(infile, O_RDONLY);
	if (fd == -1)
		goto infile_err;
	total = infile_stat.st_size;
	
	chip_start = flash->base;
	chip_end   = flash->base + flash->chip_size - 1;
	if ((dest < chip_start) || ((dest+total-1) >= chip_end))
		goto out_of_range;
	
	/* how many sectors we are writing? */
	start = dest & SECTOR_MASK;
	end = (dest+total+flash->sector_size-1) & SECTOR_MASK;
	secnr = (end - start)/(flash->sector_size);
	
	for (i = 0; i < secnr; i++) {
		printf("Erasing %s flash sector [0x%.8lX, 0x%.8lX) ... ",
			flash->name, start, start+flash->sector_size);
		ret = flash->erase_sector(start);
		if (ret) {
			printf("[FAILED]\n");
			break;
		}
		printf("[OK]\n");
		start += flash->sector_size;
	}
	
	if (ret)
		return ret;
	
	printf("Writing %s Flash memory [0x%.8lX, 0x%.8lX) ...",
		flash->name, dest, dest+total);
	fflush(stdout);
	
	while (total > 0) {
		/* try to read 4K data each time */
		buf_len = total;
		if (buf_len > PAGE_SIZE)
			buf_len = PAGE_SIZE;
		buf_len = read(fd, buf, buf_len);
		if (buf_len == -1) {
			printf(flwrite_cmd.name);
			perror(infile);
			close(fd);
			return -1;
		} else if (buf_len == 0) {
			break; /* end-of-file */
		}
		
		ret = flash->write(dest, buf_len, (u32)buf, 1);
		if (ret)
			break;
			
		dest += buf_len;
		total -= buf_len;
	} /* end of while (total > 0) */
	
	if (ret)
		printf(" [FAILED]\n");
	else
		printf(" [OK]\nTotal %lu bytes are written.\n", infile_stat.st_size);
	
	return ret;
	
out_of_range:
	printf("Error: address [0x%.8lX, 0x%.8lX) isn't within the address range"\
		" [0x%.8lX, 0x%.8lX) of current flash chip.\n",
		dest, dest+total, flash->base, flash->base+flash->chip_size);
	close(fd);
	return -1;
infile_err:
	printf(flwrite_cmd.name);
	perror(infile);
	return -1;
}

/*
 * do_flwrite_memory - write flash from target memory block
 */
static int do_flwrite_memory(u32 dest, u32 from, u32 bytenr)
{
	u32 chip_start, chip_end;
	u32 start, end;
	int secnr, i, ret;
	
	chip_start = flash->base;
	chip_end = flash->base + flash->chip_size - 1;
	if ((dest < chip_start) || ((dest+bytenr-1) >= chip_end))
		goto out_of_range;
	
	/* how many sectors we are writing? */
	start = dest & SECTOR_MASK;
	end = (dest+bytenr+flash->sector_size-1) & SECTOR_MASK;
	secnr = (end - start)/(flash->sector_size);

	for (i = 0; i < secnr; i++) {
		printf("Erasing %s flash sector [0x%.8lX, 0x%.8lX) ... ",
			flash->name, start, start+flash->sector_size);
		ret = flash->erase_sector(start);
		if (ret) {
			printf("[FAILED]\n");
			break;
		}
		printf("[OK]\n");
		start += flash->sector_size;
	}

	if (ret)
		return ret;
	
	printf("Writing %s Flash memory [0x%.8lX, 0x%.8lX) ...",
		flash->name, dest, dest+bytenr);
	fflush(stdout);
	
	ret = flash->write(dest, bytenr, from, 0);
	if (ret)
		printf(" [FAILED]\n");
	else
		printf(" [OK]\nTotal %lu bytes are written.\n", bytenr);
	
	return ret;
out_of_range:
	printf("Error: address [0x%.8lX, 0x%.8lX) isn't within the address range"\
		" [0x%.8lX, 0x%.8lX) of current flash chip.\n",
		dest, dest+bytenr, flash->base, flash->base+flash->chip_size);
	return -1;
} /* end of do_flwrite_memory(...) */

static int flwrite_run(int argc, char **argv)
{
	int size_fmt, option_index = 0, ret = 0;
	u32 to, from, size;
	char *invalid_arg;
	
	/*
	 * Step 1: reset all global option flag variables
	 * and global option value strings
	 */
	if (argc == 1)
		return do_help(&flwrite_cmd);
	to_str = from_str = size_str = infile_str = NULL;
	flwrite_subcmd = 0;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, flwrite_short_options,
				flwrite_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (flwrite_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 't':
			/* --to option */
			if (to_str)
				goto syntax_err;
			to_str = optarg;
			break;
		case 'f':
			/* --from option */
			if (from_str)
				goto syntax_err;
			from_str = optarg;
			break;
		case 's':
			/* --size option */
			if (size_str)
				goto syntax_err;
			size_str = optarg;
			break;
		case 'i':
			/* --infile option */
			if (infile_str)
				goto syntax_err;
			infile_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (size_str)
			goto syntax_err;
		size_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3 - Translate command parameters
	 */
	/* translate address argument */
	if (to_str == NULL)
		goto syntax_err;
	invalid_arg = to_str;
	ret = str2int(to_str, &to);
	if (ret)
		goto arg_err;
	
	if (infile_str) {
		if (from_str || size_str)
			goto syntax_err;
		flwrite_subcmd = FLWRITE_SUBCMD_INFILE;
	} else {
		if (from_str == NULL)
			goto syntax_err;
		invalid_arg = from_str;
		ret = str2int(from_str, &from);
		if (ret)
			goto arg_err;

		/* translate length argument */
		invalid_arg = size_str;
		size_fmt = get_memsize(size_str, &size);
		if (size_fmt == -1 || size <= 0)
			goto arg_err;
		
		if (size_fmt == WORD_FMT) {
			to   = addr32_align(to);
			from = addr32_align(from);
		} else if (size_fmt == HALFWORD_FMT) {
			to   = addr16_align(to);
			from = addr16_align(from);
		}
		flwrite_subcmd = FLWRITE_SUBCMD_MEMORY;
	}
	
	if (flash == NULL)
		goto no_flash;
	if (target->status == TARGET_STATUS_RUNNING)
		goto target_err;
	
	/*
	 * Step 4: go, go, go ...
	 */
	switch (flwrite_subcmd) {
	case FLWRITE_SUBCMD_MEMORY:
		ret = do_flwrite_memory(to, from, (size * size_fmt));
		break;
	case FLWRITE_SUBCMD_INFILE:
		ret = do_flwrite_infile(to, infile_str);
		break;
	default:
		goto syntax_err;
	}
	return ret;
	
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
no_flash:
	printf("%s: no FLASH chip is selected currently. Please use "\
		"\"flprobe\" first.\n", argv[0]);
	return -1;
target_err:
	printf("error: %s\n", err_str(-ERR_TARGET_IS_RUNNING));
	return -1;
} /* end of flwrite_run() */

cmd_t flwrite_cmd = {
	.name = "flwrite",
	.desc = "Write data into Flash memory.",
	.alias = NULL,
	.usage = "flwrite --to=addr1 --from=addr2 [[--size=]count]\n\tflwrite --to=addr1 --infile=pathname",
	.verbose = NULL,
	.long_options = flwrite_options,
	.optdesc = flwrite_optdesc,
	.non_option_args = 1,

	.run = flwrite_run,
};
