/*
 * cmdline/cmd_help.c : implements the JTAGER help command
 *
 * Copyright (C) 2003-2004, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_help.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"

#define HELP_SUBCMD_LIST   0x01
#define HELP_SUBCMD_HELP   0x02

static int help_subcmd = 0;

static char *cmdname_str = NULL; /* --cmdname option */
static char *help_short_options = "c";
static struct option help_options[] = {
	{"cmdname", required_argument, 0, 'c'},
	{0, 0, 0, 0}
};

static struct option_desc help_optdesc[] = {
	{"cmdname", "A legal command name."},
	{NULL, NULL}
};

static int print_cmd_list(void)
{
	int i = 0;
	cmd_t *cmd;
	
	cmd = cmds_tbl[0];
	while (cmd) {
		if (strcasecmp(cmd->name, "group") == 0) {
			if (i++ != 0) printf("\n");
			printf("*** %s:\n", cmd->desc);
			cmd = cmds_tbl[i];
			continue;
		}
		printf("%s\t", cmd->name);
		if (strlen(cmd->name) < 8)	
			printf("\t");
		printf("%s\n", cmd->desc);
		i++;
		cmd = cmds_tbl[i];
	};
	return 0;
}

cmd_t help_cmd;

int do_help(cmd_t *cmd)
{
	int i, j;
	struct option *opt;
	struct option_desc *optdesc;
	
	if (cmd == NULL)
		return -1;
	
	printf("USAGE:\n\t%s\n", (cmd->usage) ? cmd->usage : cmd->name);
	if (cmd->alias)
		printf("ALIAS: %s\n", cmd->alias);
	printf("DESCRIPTION:\n\t%s\n", (cmd->verbose) ? cmd->verbose : cmd->desc);
	if (cmd->long_options && cmd->optdesc) {
		printf("OPTIONS and PARAMETERS:\n");
		i = 0;
		while (1) {
			opt = &cmd->long_options[i++];
			optdesc = NULL;
			if (opt->name == NULL)
				break; /* the last option should be {0, 0, 0, 0} */
			j = 0;
			while (1) {
				optdesc = &cmd->optdesc[j++];
				if (optdesc->name == NULL)
					break;
				if (strcasecmp(opt->name, optdesc->name) == 0)
					break;
			}
			printf("\t--%s: %s\n", opt->name, optdesc ? optdesc->desc : "<NULL>");
		}
	}
	return 0;
}

static int help_run(int argc, char **argv)
{
	int ret = 0;
	int option_index = 0;
	cmd_t *cmd;
	
	/*
	 * Step 1 - reset all option flag variables and option value strings
	 */
	cmdname_str = NULL;
	help_subcmd = 0;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, help_short_options,
				help_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (help_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 'c':
			/* --cmdname option */
			if (cmdname_str)
				goto syntax_err;
			cmdname_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (cmdname_str)
			goto syntax_err;
		cmdname_str = (char *)argv[argc-1];
	}
	
	/* 
	 * Step 3 - set sub-command to be taken.
	 * NOTE: Only one sub-command can be executed.
	 */
	help_subcmd = HELP_SUBCMD_LIST;
	if (cmdname_str) {
		cmd = cmd_index(cmdname_str);
		if (cmd == NULL)
			goto unknown_cmd;
		help_subcmd = HELP_SUBCMD_HELP;
	}
	
	/* 
	 * Step 4 - go
	 */
	switch (help_subcmd) {
	case HELP_SUBCMD_LIST:
		print_cmd_list();
		break;
	case HELP_SUBCMD_HELP:
		ret = do_help(cmd);
		break;
	default:
		goto syntax_err;
	} /* end of switch (...) */
	
	return ret;
syntax_err:
	return print_syntax_error(argv[0]);
unknown_cmd:
	printf("help: unknown command: %s.\n", cmdname_str);
	return -1;
} /* end of help_run() */

cmd_t help_cmd = {
	.name = "help",
	.desc = "Show all JTAGER commands list, or verbose help info",
	.alias = "ls",
	.usage = "help/ls [[--cmdname=]cmdname]",
	.verbose = NULL,
	.long_options = help_options,
	.optdesc = help_optdesc,
	.non_option_args = 1,

	.run = help_run,
};
