/*
 * cmdline/cmd_ice.c : implements the JTAGER ice command.
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_ice.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"
#include "target.h"

#define ICE_SUBCMD_LIST 0x01
#define ICE_SUBCMD_READ_ALL 0x02
#define ICE_SUBCMD_READ 0x04
#define ICE_SUBCMD_WRITE 0x08

static int ice_subcmd = 0;

static int list_flag = 0; /* --list option */
static char *name_str = NULL; /* --name option */
static char *value_str = NULL; /* --value option */
static const char *ice_short_options = "nv";
static struct option ice_options[] = {
	{"list", no_argument, &list_flag, 1},
	{"name", required_argument, 0, 'n'},
	{"value", required_argument, 0, 'v'},
	{0, 0, 0, 0}
};

static struct option_desc ice_optdesc[] = {
	{"list", "Print a list of all available ICE register names."},
	{"name", "The name of EmbeddedICE-RT logic registers."},
	{"value", "The new value to be written to the ICE register."},
	{NULL, NULL}
};

static int do_ice_list(void)
{
	ice_register_t *reg;
	
	reg = target->ice_regs;
	if (reg->name == NULL) {
		printf("The current target no available ICE registers.\n");
		return -1;
	}
	
	printf("The EmbeddedICE-RT logic registers for ARM7TDMI target:\n");
	while (reg->name) {
		printf("\t%s:\t%s\n", reg->name, reg->desc);
		reg++;
	}
	return 0;
} /* end of do_ice_list() */

static int do_ice_read_all(void)
{
	int ret = 0;
	u32 value = 0;
	ice_register_t *reg;

	printf("Reading all ICE registers:\n");
	reg = target->ice_regs;
	while (reg && reg->name) {
		printf("\t%s:\t", reg->name);
		ret = t_op->ice_read(reg->addr, &value);
		if (ret) {
			printf("FAILED\nError:%s", err_str(ret));
			break;
		} else {
			printf("0x%.8lX\n", value);
		}
		reg++;
	}
	return ret;
} /* end of do_ice_read_all() */

static int do_ice_read(const ice_register_t *reg)
{
	u32 value;
	int ret;
	
	printf("Reading ICE %s ... ", reg->desc);
	ret = t_op->ice_read(reg->addr, &value);
	if (ret)
		printf("[FAILED]\nError:%s", err_str(ret));
	else
		printf("0x%.8lX\n", value);

	return ret;
} /* end of do_ice_read(...) */

static int do_ice_write(const ice_register_t *reg, u32 newval)
{
	int ret;

	printf("Setting new value 0x%lX to ICE %s ... ", newval, reg->desc);
	ret = t_op->ice_write(reg->addr, newval);
	if (ret)
		printf("FAILED\nError:%s", err_str(ret));
	else
		printf("OK\n");
	return ret;
} /* end of do_ice_write(...) */

static int ice_run(int argc, char **argv)
{
	int option_index = 0, ret = 0;
	u32 newval;
	ice_register_t *reg = NULL;
	char *invalid_arg = NULL;

	/*
	 * Step 1 - reset all option flag variables and option value strings
	 */
	list_flag = 0;
	name_str = value_str = NULL;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, ice_short_options,
				ice_options, &option_index);

		/* detec the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (ice_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 'n':
			/* --name option */
			if (name_str)
				goto syntax_err;
			name_str = optarg;
			break;
		case 'v':
			/* --value option */
			if (value_str)
				goto syntax_err;
			value_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */

	/* Try to get non-option argument */
	if (optind < argc) {
		if (value_str)
			goto syntax_err;
		value_str = (char *)argv[argc-1];
	}

	/* 
	 * Step 3 - set sub-command to be taken.
	 * NOTE: Only one sub-command can be executed.
	 */
	ice_subcmd = 0;
	if (argc == 1)
		ice_subcmd = ICE_SUBCMD_READ_ALL;
		
	if (list_flag) {
		if (ice_subcmd)
			goto syntax_err;
		if (argc != 2)
			goto syntax_err;
		ice_subcmd = ICE_SUBCMD_LIST;
	}

	/*
	 * Read or write the specified ICE register.
	 */
	if (name_str) {
		reg = ice_reg_index(name_str);
		if (reg == NULL)
			goto unknown_reg;
		if (value_str == NULL) {
			ice_subcmd = ICE_SUBCMD_READ;
		} else {
			/* translate new value string */
			invalid_arg = value_str;
			ret = str2int(value_str, &newval);
			if (ret)
				goto arg_err;
			ice_subcmd = ICE_SUBCMD_WRITE;
		}
	}

	/*
	 * Step 4 - go
	 */
	switch (ice_subcmd) {
	case ICE_SUBCMD_LIST:
		ret = do_ice_list();
		break;
	case ICE_SUBCMD_READ_ALL:
		ret = do_ice_read_all();
		break;
	case ICE_SUBCMD_READ:
		ret = do_ice_read(reg);
		break;
	case ICE_SUBCMD_WRITE:
		ret = do_ice_write(reg, newval);
		break;
	default:
		goto syntax_err;
	} /* end of switch (ice_subcmd) */
	
	return ret;
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
unknown_reg:
	printf("%s: unknown ICE register name: %s.\n", argv[0], name_str);
	printf("Please use \"ice --list\" command to get a "\
		"list of available ICE register names.\n");
	return -1;
} /* end of ice_run() */

cmd_t ice_cmd = {
	.name = "ice",
	.desc = "Read/write the registers of EmbeddedICE-RT logic",
	.alias = NULL,
	.usage = "ice [--list]\n\tice --name=regname [[--value=]newvalue]",
	.verbose = "This command enables us to read or write all the registers"\
		"of the\n\tEmbeddedICE-RT logic in the target CPU core.",
	.long_options = ice_options,
	.optdesc = ice_optdesc,
	.non_option_args = 1,

	.run = ice_run,
};
