/*
 * cmdline/cmd_idcode.c : implements the JTAGER idcode command
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_idcode.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "jtager.h"
#include "jtag.h"
#include "types.h"

#define IDCODE_MANUFACTUER_MASK 	0x00000FFE
#define IDCODE_PARTNUMBER_MASK   	0x0FFFF000
#define IDCODE_VERION_MASK 		0xF0000000

#define IDCODE_MANUFACTURER(id) \
	(((id) & IDCODE_MANUFACTUER_MASK) >> 1)

#define IDCODE_PART_NUMBER(id) \
	(((id) & IDCODE_PARTNUMBER_MASK) >> 12)

#define IDCODE_VERSION(id) \
	(((id) & IDCODE_VERION_MASK) >> 28)

/* bits 11-1 of the Device Identification Register */
typedef struct manufacturer_struct {
	u32 id; /* bit[11:1] of IDCODE */
	char *name;
}manufacturer_t;

/* All manufacturers known to us */
static manufacturer_t manufacturers[] = {
	{0x07, "HITACHI"},
	{0x09, "INTEL"},
	{0x15, "PHILIPS Semi"},
	{0x1f, "ATMEL"},
	{0x34, "CYPRESS"},
	{0x35, "DEC"},
	{0x49, "XILINX"},
	{0xbf, "BROADCOM"},
	{0x150, "BROADCOM"},
	{0x787, "SAMSUNG"},
	{0x4e, "SAMSUNG"},
	/* the last one is an invalid one, its name must be NULL */
	{0x00, NULL}
};

static manufacturer_t *manufacturer_index(u32 manu_id)
{
	manufacturer_t * current;
	int i = 0, found = 0;

	while(1) {
		current = &manufacturers[i];
		if (current->name == NULL)
			break;
		if (current->id == manu_id) {
			found = 1;
			break;
		}
		i ++;
	}

	if (found)
		return current;
	else
		return NULL;
} /* end of manufacturer_index(...) */

static int idcode_run(int argc, char **argv)
{
	u32 idcode;
	u32 manu_id;
	u32 part_num;
	u32 version;
	manufacturer_t *manu;
	int retval;

	/* halt command needs no params */
	if (argc != 1)
		return print_syntax_error(argv[0]);
	
	/* go */
	printf("Reading the JTAG Device ID code ... ");
	retval = jtag_idcode(&idcode);
	if (retval) {
		printf("[FAILED]\nError:%s", err_str(retval));
		return retval;
	}

	printf("[OK]\nDevice ID = 0x%.8lX\n", idcode);
	printf("\tbit[0] = 1, always be 1, required by IEEE Std 1149.1\n");
	printf("\tManufacturer (bit[11:1]): ");
	manu_id = IDCODE_MANUFACTURER(idcode);
	printf("0x%.3lX - ", manu_id);
	manu = manufacturer_index(manu_id);
	if (manu)
		printf("%s\n", manu->name);
	else
		printf("unknown\n");

	part_num = IDCODE_PART_NUMBER(idcode);
	printf("\tPart number (bit[27:12]): 0x%.4lX\n", part_num);

	version = IDCODE_VERSION(idcode);
	printf("\tVersion (bit[31:28])    : 0x%.1lX\n", version);
	
	return 0;
} /* end of idcode_run() */

cmd_t idcode_cmd = {
	.name = "idcode",
	.desc = "Read the jtag IDCODE of your target",
	.alias = NULL,
	.usage = NULL,
	.verbose = "Read the JTAG IDCODE of your target CPU. Please refer " \
		"to the\n\tBSDL file of your specific target CPU for the " \
		"IDCODE value of\n\tyour specific target CPU.",
	.long_options = NULL,
	.optdesc = NULL,
	.non_option_args = 0,

	.run = idcode_run,
};
