/*
 * cmd_memcmp.c : implements the memcmp command
 *
 * Copyright (C) 2003-2004, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_memcmp.c,v 1.1.1.1 2004/09/26 03:59:44 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"
#include "target.h"

static char *base1_str = NULL;  /* --base1 option */
static char *base2_str = NULL;  /* --base2 option */
static char *size_str  = NULL;  /* --size option */
static char *memcmp_short_options = "12s";
static struct option memcmp_options[] = {
	{"base1", required_argument, 0, '1'},
	{"base2", required_argument, 0, '2'},
	{"size", required_argument, 0, 's'},
	{0, 0, 0, }
};

static struct option_desc memcmp_optdesc[] = {
	{"base1", "The base address of the first memory block."},
	{"base2", "The base address of the second memory block."},
	SIZE_OPTDESC,
	{NULL, NULL}
};

cmd_t memcmp_cmd;

/**
 * do_memcmp - Compare two memory blocks
 */
static int do_memcmp(u32 base1, u32 base2, u32 bytenr)
{
	int ret = 0, i = 0;
	u8 tmp1, tmp2;
	
	printf("%s: [0x%.8lX, 0x%.8lX) == [0x%.8lX, 0x%.8lX) ? ...",
		memcmp_cmd.name, base1, base1+bytenr, base2, base2+bytenr);
	fflush(stdout);
	
	/* compare two memory blocks */
	while (i < bytenr) {
		tmp1 = read8(base1);
		tmp2 = read8(base2);
		if (tmp1 != tmp2) {
			ret = -1; /* different */
			break;
		}
		base1++;
		base2++;
		i++;
		if ((i % PAGE_SIZE) == 0) {
			printf(".");
			fflush(stdout);
		}
	}
	
	if (ret == 0) {
		printf(" [YES]\n");
	} else {
		printf(" [NO]\n");
		printf("byte@(0x%.8lX) (=0x%.2X) != byte@(0x%.8lX) (=0x%.2X)\n",
			base1, read8(base1), base2, read8(base2));
	}
	
	return ret;
}

static int memcmp_run(int argc, char **argv)
{
	int size_fmt, option_index = 0, ret = 0;
	u32 base1, base2, size;
	char *invalid_arg;
	
	/*
	 * Step 1: reset all global option flag variables
	 * and global option value strings
	 */
	if (argc == 1)
		return do_help(&memcmp_cmd);
	base1_str = base2_str = size_str = NULL;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, memcmp_short_options,
				memcmp_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (memcmp_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case '1':
			/* --base1 option */
			if (base1_str)
				goto syntax_err;
			base1_str = optarg;
			break;
		case '2':
			/* --base2 option */
			if (base2_str)
				goto syntax_err;
			base2_str = optarg;
			break;
		case 's':
			/* --size option */
			if (size_str)
				goto syntax_err;
			size_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (size_str)
			goto syntax_err;
		size_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3 - Translate command parameters
	 */
	/* translate address argument */
	if (base1_str == NULL)
		goto syntax_err;
	if (base2_str == NULL)
		goto syntax_err;
	invalid_arg = base1_str;
	ret = str2int(base1_str, &base1);
	if (ret)
		goto arg_err;
	invalid_arg = base2_str;
	ret = str2int(base2_str, &base2);
	if (ret)
		goto arg_err;
	
	/* translate length argument */
	invalid_arg = size_str;
	size_fmt = get_memsize(size_str, &size);
	if (size_fmt == -1 || size <= 0)
		goto arg_err;
	
	if (size_fmt == WORD_FMT) {
		base1 = addr32_align(base1);
		base2 = addr32_align(base2);
	} else if (size_fmt == HALFWORD_FMT) {
		base1 = addr16_align(base1);
		base2 = addr16_align(base2);
	}
	
	if (target->status == TARGET_STATUS_RUNNING)
		goto target_err;
	
	/*
	 * Step 4: go, go, go ...
	 */
	return do_memcmp(base1, base2, (size * size_fmt));
	
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
target_err:
	printf("error: %s\n", err_str(-ERR_TARGET_IS_RUNNING));
	return -1;
} /* end of memcmp_run() */

cmd_t memcmp_cmd = {
	.name = "memcmp",
	.desc = "Compare the contents of two memory blocks.",
	.alias = NULL,
	.usage = "memcmp --base1=addr1 --base2=addr2 [[--size=]count]",
	.verbose = NULL,
	.long_options = memcmp_options,
	.optdesc = memcmp_optdesc,
	.non_option_args = 1,

	.run  = memcmp_run,
};
