/*
 * cmd_memcpy.c : implements the memcpy command
 *
 * Copyright (C) 2003 - 2004, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_memcpy.c,v 1.1.1.1 2004/09/26 03:59:45 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"
#include "target.h"

static char *to_str = NULL; /* --to option */
static char *from_str = NULL; /* --from option */
static char *size_str = NULL; /* --size option */
static char *memcpy_short_options = "tfs";
static struct option memcpy_options[] = {
	{"to", required_argument, 0, 't'},
	{"from", required_argument, 0, 'f'},
	{"size", required_argument, 0, 's'},
	{0, 0, 0, 0}
};

static struct option_desc memcpy_optdesc[] = {
	{"to", "The base address of the destination memory block."},
	{"from", "The base address of the source memory block."},
	SIZE_OPTDESC,
	{NULL, NULL}
};

cmd_t memcpy_cmd;

/**
 * do_memcpy - Copy memory blocks
 * @sizefmt: the format of size: BYTE ? HALFWORD ? WORD
 * @size: If sizefmt == BYTE_FMT, it means how many bytes.
 *           If sizefmt == HALFWORD_FMT, it means how many halfwords.
 *           If sizefmt == WORD_FMT, it means how many words.
 */
static int do_memcpy(u32 to, u32 from, u32 bytenr)
{
	u32 total = bytenr;
	u8 val;

	printf("%s: [0x%.8lX, 0x%.8lX) => [0x%.8lX, 0x%.8lX) ...",
		memcpy_cmd.name, from, from+bytenr, to, to+bytenr);
	fflush(stdout);
	
	while (total--) {
		val = read8(from);
		write8(val, to);
		to++;
		from++;
		if (((bytenr - total) % PAGE_SIZE) == 0) {
			printf(".");
			fflush(stdout);
		}
	}
	
	printf(" [OK]\nTotal %lu bytes are copied.\n", bytenr);
	return 0;
}

static int memcpy_run(int argc, char **argv)
{
	int size_fmt, option_index = 0, ret = 0;
	u32 to, from, size, bytenr;
	char *invalid_arg = NULL;
	
	/*
	 * Step 1: reset all option flag variables and option value strings
	 */
	if (argc == 1)
		return do_help(&memcpy_cmd);
	to_str = from_str = size_str = NULL;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, memcpy_short_options,
				memcpy_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (memcpy_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 't':
			/* --to option */
			if (to_str)
				goto syntax_err;
			to_str = optarg;
			break;
		case 'f':
			/* --from option */
			if (from_str)
				goto syntax_err;
			from_str = optarg;
			break;
		case 's':
			/* --size option */
			if (size_str)
				goto syntax_err;
			size_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (size_str)
			goto syntax_err;
		size_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3 - Translate command parameters
	 */
	/* translate address argument */
	if (to_str == NULL)
		goto syntax_err;
	if (from_str == NULL)
		goto syntax_err;
	invalid_arg = to_str;
	ret = str2int(to_str, &to);
	if (ret)
		goto arg_err;
	invalid_arg = from_str;
	ret = str2int(from_str, &from);
	if (ret)
		goto arg_err;
	
	/* translate length argument */
	invalid_arg = size_str;
	size_fmt = get_memsize(size_str, &size);
	if (ret == -1 || size <= 0)
		goto arg_err;
	
	bytenr = size * size_fmt;
	if (size_fmt == HALFWORD_FMT) {
		to = addr16_align(to);
		from = addr16_align(from);
	} else if (size_fmt == WORD_FMT) {
		to = addr32_align(to);
		from = addr32_align(from);
	}
	
	/*
	 * Check whether the dest address space and the source
	 * address space are interleaved.
	 */
	if ((to == from)
		|| ((to > from) && (from+bytenr) > to)
		|| ((to < from) && (to+bytenr) > from))
		goto space_interleaved;
	
	if (target->status == TARGET_STATUS_RUNNING)
		goto target_err;
	
	/*
	 * Step 4: go, go, go ...
	 */
	return do_memcpy(to, from, bytenr);
	
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
space_interleaved:
	printf("%s: the source address space and the destination "\
		"address space are interleaved.\n", argv[0]);
	return -1;
target_err:
	printf("error: %s\n", err_str(-ERR_TARGET_IS_RUNNING));
	return -1;
} /* end of memcpy_run() */

cmd_t memcpy_cmd = {
	.name = "memcpy",
	.desc = "Copy memory block from one to another.",
	.alias = NULL,
	.usage = "memcpy --to=addr1 --from=addr2 [[--size=]count]",
	.verbose = NULL,
	.long_options    = memcpy_options,
	.optdesc = memcpy_optdesc,
	.non_option_args = 1,

	.run  = memcpy_run,
};
