/*
 * cmd_memdump.c : implements the memdump command
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_memdump.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"
#include "target.h"

#define MEMDUMP_SUBCMD_STDOUT  0x01
#define MEMDUMP_SUBCMD_OUTFILE 0x02

static int memdump_subcmd = 0;

static char *base_str = NULL; /* --base option */
static char *size_str = NULL; /* --size option */
static char *outfile_str = NULL; /* --outfile option */
static char *memdump_short_options = "bso";
static struct option memdump_options[] = {
	{"base", required_argument, 0, 'b'},
	{"size", required_argument, 0, 's'},
	{"outfile", required_argument, 0, 'o'},
	{0, 0, 0, 0}
};

static struct option_desc memdump_optdesc[] = {
	{"base", "The base address of the memory block."},
	SIZE_OPTDESC,
	{"outfile", "The pathname of the output file"},
	{NULL, NULL}
};
static const char *more_prompt = "Do you want to read more?";

cmd_t memdump_cmd;

/**
 * isprint - is a printable character ?
 */
static int isprint(int c)
{
	if ((c >= 0x20) && (c <= 0x7E))
		return 1;
	else
		return 0;
}

/**
 * output_printchars - output a memory line (16 bytes) in printable chars mode
 * @buf: pointer to a char array which contains 16 chars.
 */
static void output_printchars(char *buf, char *showbuf)
{
	int i = 0;
	
	*showbuf++ = '|';
	while (i < 16) {
		if (isprint(buf[i]))
			*showbuf++ = (char)buf[i];
			//sprintf(showbuf, "%c", buf[i]);
		else
			//sprintf(showbuf, ".");
			*showbuf++ = '.';
		i++;
		//showbuf++;
	}
	*showbuf++ = '|';
	*showbuf++ = '\n';
}

static int do_memdump_outfile(u32 base, u32 bytenr, char *outfile)
{
	FILE *fp;
	u32 i = 0;
	u8 tmp;
	
	fp = fopen(outfile, "w");
	if (fp == NULL) {
		perror(outfile);
		return -1;
	}
	
	printf("memdump: Dump the memory block [0x%.8lX, 0x%.8lX) to the host binary file "
		"\"%s\" ...", base, base+bytenr, outfile);
	fflush(stdout);
	
	while (i < bytenr) {
		tmp = read8(base++);
		fwrite(&tmp, 1, 1, fp);
		i++;
		if ((i % PAGE_SIZE) == 0) {
			printf(".");
			fflush(stdout);
		}
	}
	
	printf(" [OK]\nTotal %lu bytes are dumped.\n", bytenr);
	fclose(fp);
	return 0;
}

static int do_memdump(u32 base, int sizefmt, u32 size)
{
	u8 buf[16], *ptr8;
	u16 *ptr16;
	u32 *ptr32;
	int i, j, line_cnt;
	char showbuf[80], *showptr;
	
	/* Each memory line shows: 4 words, 8 halfwords or 16 bytes
	 * - sizefmt = 4 (WORD_FMT)    : line_cnt = 4
	 * - sizefmt = 2 (HALFWORD_FMT): line_cnt = 8
	 * - sizefmt = 1 (BYTE_FMT)    : line_cnt = 16
	 */
	line_cnt = 16/sizefmt;
	showbuf[79] = '\0';
	i = 0;
	while (i < size) {
		if ((i%line_cnt) == 0) {
			/* start a new line */
			memset(buf, 0, 16);
			memset(showbuf, ' ', 79);
			ptr32 = (u32 *)buf;
			ptr16 = (u16 *)buf;
			ptr8  = (u8 *)buf;
			showptr = showbuf;
			j = sprintf(showptr, "%.8lX: ", base);
			showptr += j;
			*showptr = ' '; /* remove the null terminator added by sprintf */
		}
		if ((((i+line_cnt)/line_cnt) % 20) == 0) {
			if (ask_yesno(more_prompt) == 0)
				break;
		}
		
		if (sizefmt == WORD_FMT) {
			*ptr32 = read32(base);
			j = sprintf(showptr, "%.8lX ", *ptr32);
			base += 4;
			*ptr32++;
		} else if (sizefmt == HALFWORD_FMT) {
			*ptr16 = read16(base);
			j = sprintf(showptr, "%.4X ", *ptr16);
			base += 2;
			*ptr16++;
		} else {
			*ptr8 = read8(base);
			j = sprintf(showptr, "%.2X ", *ptr8);
			base++;
			*ptr8++;
		}
		showptr += j;
		*showptr = ' '; /* remove the null terminator added by sprintf */
		
		if ((i+1) % (line_cnt/2) == 0)
			showptr++;
		
		if (((i+1) % line_cnt == 0) || ((i+1) == size)) {
			output_printchars(buf, showbuf+60);
			printf(showbuf);
		}
		i++;
	} /* end of while (i < size) */
	
	return 0;
}

static int memdump_run(int argc, char **argv)
{
	int size_fmt, option_index = 0, ret;
	u32 base, size;
	char *invalid_arg = NULL;
	
	/*
	 * Step 1: reset all global option flag variables and
	 * global option value strings
	 */
	if (argc == 1)
		return do_help(&memdump_cmd);
	base_str = size_str = outfile_str = NULL;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, memdump_short_options,
				memdump_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (memdump_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 'b':
			/* --base option */
			if (base_str)
				goto syntax_err;
			base_str = optarg;
			break;
		case 's':
			/* --size option */
			if (size_str)
				goto syntax_err;
			size_str = optarg;
			break;
		case 'o':
			/* --outfile option */
			if (outfile_str)
				goto syntax_err;
			outfile_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (base_str)
			goto syntax_err;
		base_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3: translate option parameters and non-option paramters
	 */
	/* translate address argument */
	if (base_str == NULL)
		goto syntax_err;
	invalid_arg = base_str;
	ret = str2int(base_str, &base);
	if (ret)
		goto arg_err;
	
	/* translate size argument */
	invalid_arg = size_str;	
	size_fmt = get_memsize(size_str, &size);
	if (size_fmt == -1 || size <= 0)
		goto arg_err;
	
	if (size_fmt == WORD_FMT)
		base = addr32_align(base);
	else if (size_fmt == HALFWORD_FMT)
		base = addr16_align(base);
	
	if (outfile_str)
		memdump_subcmd = MEMDUMP_SUBCMD_OUTFILE;
	else
		memdump_subcmd = MEMDUMP_SUBCMD_STDOUT;
	
	if (target->status == TARGET_STATUS_RUNNING)
		goto target_err;
	
	/*
	 * Step 4: go, go, go ...
	 */
	if (memdump_subcmd == MEMDUMP_SUBCMD_STDOUT)
		return do_memdump(base, size_fmt, size);
	else
		return do_memdump_outfile(base, (size * size_fmt), outfile_str);
	
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
target_err:
	printf("error: %s\n", err_str(-ERR_TARGET_IS_RUNNING));
	return -1;
} /* end of memdump_run(...) */

cmd_t memdump_cmd = {
	.name = "memdump",
	.desc = "Display the contents of a memory block.",
	.alias = NULL,
	.usage = "memdump [--base=]addr [--size=count] [--outfile=pathname]",
	.verbose = NULL,
	.long_options    = memdump_options,
	.optdesc = memdump_optdesc,
	.non_option_args = 1,

	.run  = memdump_run,
};
