/*
 * cmd_memset.c : implements the memset command
 *
 * Copyright (C) 2003 - 2004, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_memset.c,v 1.1.1.1 2004/09/26 03:59:47 zhanrk Exp $ */

#include <unistd.h>
#define _GNU_SOURCE /* for getline() function */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "jtager.h"
#include "target.h"

#define MEMSET_SUBCMD_AUTOINC 0x01
#define MEMSET_SUBCMD_SETBLK  0x02
#define MEMSET_SUBCMD_INFILE  0x04

static int memset_subcmd = 0;

static char *base_str = NULL; /* --base option */
static char *size_str = NULL; /* --size option */
static char *value_str = NULL; /* --value option */
static char *infile_str = NULL; /* --infile option */
static char *memset_short_options = "bsvi";
static struct option memset_options[] = {
	{"base", required_argument, 0, 'b'},
	{"size", required_argument, 0, 's'},
	{"value", required_argument, 0, 'v'},
	{"infile", required_argument, 0, 'i'},
	{0, 0, 0, 0}
};

static struct option_desc memset_optdesc[] = {
	{"base", "The base address of the memory block."},
	SIZE_OPTDESC,
	{"value", "The new value to be set"},
	{"infile", "The pathname of the input data file"},
	{NULL, NULL}
};

cmd_t memset_cmd;

/**
 * do_memset_autoinc - Set target memory by auto-incremented method
 */
static int do_memset_autoinc(u32 base, int sizefmt)
{
	int ret, len;
	u32 newval;
	char *newval_str;

	printf("press 'q' or 'Q' or 'ENTER' to exit.\n");
	while (1) {
		printf("0x%.8lX: ", base);
		if (sizefmt == WORD_FMT)
			printf("%.8lX", read32(base));
		else if (sizefmt == HALFWORD_FMT)
			printf("%.4X", read16(base));
		else
			printf("%.2X", read8(base));
		printf(" ? ");
		
		newval_str = NULL;
		len = 0;
		len = getline(&newval_str, &len, stdin);
		
		/* quit ? */
		if ((ret = strcasecmp(newval_str, "\n")) == 0 ||
			((ret = strcasecmp(newval_str, "q\n")) == 0) ||
			((ret = strcasecmp(newval_str, "Q\n")) == 0) ||
			((ret = strcasecmp(newval_str, "quit\n")) == 0)) {
			free(newval_str); /* free the buffer allocated by getline() */
			break;
		}
		
		if (newval_str[len-1] == '\n')
			newval_str[len-1] = '\0';
		ret = str2int(newval_str, &newval);
		if (ret) {
			printf("error: invalid new value: %s\n", newval_str);
			continue;
		}

		/* set new value, and echo it */
		printf("0x%.8lX: ", base);
		if (sizefmt == WORD_FMT) {	
			write32(newval, base);
			printf("%.8lX\n", read32(base));
			base += 4;
		} else if (sizefmt == HALFWORD_FMT) {
			write16((newval & 0xFFFF), base);
			printf("%.4X\n", read16(base));
			base += 2;
		} else {
			write8((newval & 0xFF), base);
			printf("%.2X\n", read8(base));
			base++;
		}
		
		if (newval_str)
			free(newval_str); /* free the buffer allocated by getline() */
	} /* end of while (1) */
	
	return 0;
} /* end of do_memset_autoinc() */

static int do_memset_block(u32 base, int sizefmt, u32 size, u32 value)
{
	u32 bytenr = size * sizefmt;
	
	if (bytenr > 4)
		printf("%s: The memory block [0x%.8lX, 0x%.8lX) is filled with new value: 0x%lX\n",
			memset_cmd.name, base, base+bytenr, value);
	
	switch (sizefmt) {
	case WORD_FMT:
		while (size--) {
			write32(value, base);
			base += 4;
		}
		break;
	case HALFWORD_FMT:
		while (size--) {
			write16(value, base);
			base += 2;
		}
		break;
	case BYTE_FMT:
	default:
		while (size--) {
			write8(value, base);
			base++;
		}
		break;
	} /* end of switch (...) */
	
	if (bytenr > 4)
		printf("Total %lu bytes are filled.\n", bytenr);
	return 0;
} /* end of do_memset_block() */

static int do_memset_infile(u32 base, char *infile)
{
	struct stat infile_stat;
	int fd, ret;
	unsigned long total;
	ssize_t buf_len;
	u8 buf[PAGE_SIZE];
	
	ret = stat(infile, &infile_stat);
	if (ret)
		goto infile_err;
	
	fd = open(infile, O_RDONLY);
	if (fd == -1)
		goto infile_err;
	total = infile_stat.st_size;
	printf("%s: Fill the memory block [0x%.8lX, 0x%.8lX] with the file "\
		"\"%s\" ...", memset_cmd.name, base, base+total-1, infile);
	fflush(stdout);
	
	while (total > 0) {
		/* try to read 4K data each time */
		buf_len = total;
		if (buf_len > PAGE_SIZE)
			buf_len = PAGE_SIZE;
		buf_len = read(fd, buf, buf_len);
		if (buf_len == -1) {
			printf(memset_cmd.name);
			perror(infile);
			goto out;
		} else if (buf_len == 0) {
			break; /* end-of-file */
		}
		
		t_op->mem_write8(buf, base, buf_len);
			
		base += buf_len;
		total -= buf_len;
		printf(".");
		fflush(stdout);
	} /* end of while (total > 0) */

out:
	if (ret == 0)
		printf(" [OK]\nTotal %lu bytes are written.\n", infile_stat.st_size);
	else
		printf(" [FAILED]\n");
	close(fd);
	return ret;
infile_err:
	printf(memset_cmd.name);
	perror(infile);
	return -1;
} /* end of do_memset_infile() */

static int memset_run(int argc, char **argv)
{
	int size_fmt, option_index = 0, ret = 0;
	u32 base, size, value;
	char *invalid_arg;
	
	/*
	 * Step 1: reset all global option flag variables
	 * and global option value strings
	 */
	if (argc == 1)
		return do_help(&memset_cmd);
	base_str = size_str = NULL;
	value_str = infile_str = NULL;
	memset_subcmd = 0;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, memset_short_options,
				memset_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (memset_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 'b':
			/* --base option */
			if (base_str)
				goto syntax_err;
			base_str = optarg;
			break;
		case 's':
			/* --size option */
			if (size_str)
				goto syntax_err;
			size_str = optarg;
			break;
		case 'v':
			/* --value option */
			if (value_str)
				goto syntax_err;
			value_str = optarg;
			break;
		case 'i':
			/* --infile option */
			if (infile_str)
				goto syntax_err;
			infile_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (base_str)
			goto syntax_err;
		base_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3 - Translate command parameters
	 */
	if (base_str == NULL)
		goto syntax_err;
	
	/* first, set format */
	if (size_str == NULL && value_str == NULL && infile_str == NULL) {
		/* Auto-incremented way ...
		 * so get the memsize format for base address option.
		 */
		size_fmt = get_memsize(base_str, NULL);
		
		memset_subcmd = MEMSET_SUBCMD_AUTOINC;
	} else if (size_str || value_str) {
		if (infile_str)
			goto syntax_err;
		invalid_arg = size_str;
		size_fmt = get_memsize(size_str, &size);
		if (size_fmt == -1 || size <= 0)
			goto arg_err;
		
		value = 0; /* default value is 0 */
		if (value_str) {
			invalid_arg = value_str;
			ret = str2int(value_str, &value);
			if (ret)
				goto arg_err;
		}
		
		memset_subcmd = MEMSET_SUBCMD_SETBLK;
	} else {
		size_fmt = BYTE_FMT;
		memset_subcmd = MEMSET_SUBCMD_INFILE;
	}

	/* translate base address */
	invalid_arg = base_str;
	ret = str2int(base_str, &base);
	if (ret)
		goto arg_err;
	
	if (size_fmt == WORD_FMT)
		base = addr32_align(base);
	else if (size_fmt == HALFWORD_FMT)
		base = addr16_align(base);
	
	if (target->status == TARGET_STATUS_RUNNING)
		goto target_err;
	
	/*
	 * Step 4: go, go, go ...
	 */
	switch (memset_subcmd) {
	case MEMSET_SUBCMD_AUTOINC:
		ret = do_memset_autoinc(base, size_fmt);
		break;
	case MEMSET_SUBCMD_SETBLK:
		ret = do_memset_block(base, size_fmt, size, value);
		break;
	case MEMSET_SUBCMD_INFILE:
		ret = do_memset_infile(base, infile_str);
		break;
	default:
		goto syntax_err;
	}
	
	return ret;
	
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
target_err:
	printf("error: %s\n", err_str(-ERR_TARGET_IS_RUNNING));
	return -1;
} /* end of memset_run() */

cmd_t memset_cmd = {
	.name = "memset",
	.desc = "Modify the value of a memory unit.",
	.alias = NULL,
	.usage = "memset [--base=]addr\n\tmemset [--base=]addr --size=count --value=newval"
		 "\n\tmemset [--base=]addr --infile=pathname",
	.verbose = NULL,
	.long_options = memset_options,
	.optdesc = memset_optdesc,
	.non_option_args = 1,

	.run = memset_run
};
