/*
 * cmd_memtest.c : implements the memtest command
 *
 * Copyright (C) 2003 - 2004, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_memtest.c,v 1.1.1.1 2004/09/26 03:59:47 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"
#include "target.h"

static char *base_str = NULL; /* --base option */
static char *size_str = NULL; /* --size option */
static char *memtest_short_options = "bs";
static struct option memtest_options[] = {
	{"base", required_argument, 0, 'b'},
	{"size", required_argument, 0, 's'},
	{0, 0, 0, 0}
};

static struct option_desc memtest_optdesc[] = {
	{"base", "The base address of the memory block."},
	SIZE_OPTDESC,
	{NULL, NULL}
};

cmd_t memtest_cmd;

static int do_memtest(u32 base, u32 bytenr)
{
	u8 oldval, newval;
	int i = 0, ret;

	if (bytenr == 0)
		return 0;

	printf("%s: testing the memory block [0x%lX, 0x%lX] ...",
		memtest_cmd.name, base, base+bytenr-1);
	fflush(stdout);
	while (i < bytenr) {
		oldval = read8(base);

		write8(0x55, base);
		newval = read8(base);
		if (newval != 0x55) {
			ret = -1;
			break;
		}
		
		write8(0xAA, base);
		newval = read8(base);
		if (newval != 0xAA) {
			ret = -1;
			break;
		}
		
		/* write original value */
		write8(oldval, base);
		base++;
		i++;
		if ((i % PAGE_SIZE) == 0) {
			putc('.', stdout);
			fflush(stdout);
		}
	}
	if (ret == -1) {
		write8(oldval, base); /* try to write back the old value */
		printf(" [FAILED]\n");
		printf("byte@0x%.8lX is not writeable.\n", base);
	} else {
		printf(" [OK]\n");
	}
	return ret;
}

static int memtest_run(int argc, char **argv)
{
	int option_index = 0;
	int ret, size_fmt;
	u32 base, size;
	char *invalid_arg = NULL;

	if (argc == 1)
		return do_help(&memtest_cmd);
	
	/*
	 * Step 1: reset all global option flag variables and
	 * global option value strings
	 */
	base_str = size_str = NULL;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, memtest_short_options,
				memtest_options, &option_index);

		/* detect the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (memtest_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 'b':
			/* --base option */
			if (base_str)
				goto syntax_err;
			base_str = optarg;
			break;
		case 's':
			/* --size option */
			if (size_str)
				goto syntax_err;
			size_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (base_str)
			goto syntax_err;
		base_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3: translate option parameters and non-option paramters
	 */
		
	/* translate address argument */
	if (base_str == NULL)
		goto syntax_err;
	invalid_arg = base_str;
	ret = str2int(base_str, &base);
	if (ret)
		goto arg_err;
	
	/* translate size argument */
	invalid_arg = size_str;	
	size_fmt = get_memsize(size_str, &size);
	if (size_fmt == -1 || size <= 0)
		goto arg_err;
	
	if (target->status == TARGET_STATUS_RUNNING)
		goto target_err;
	
	/*
	 * Step 4: go, go, go ...
	 */
	return do_memtest(base, (size * size_fmt));
	
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
target_err:
	printf("error: %s\n", err_str(-ERR_TARGET_IS_RUNNING));
	return -1;
} /* end of memtest_run(...) */

cmd_t memtest_cmd = {
	.name = "memtest",
	.desc = "Test if a memory block is fully readable and writeable.",
	.alias = NULL,
	.usage = "memtest [--base=]addr [--size=count]",
	.verbose = NULL,
	.long_options = memtest_options,
	.optdesc = memtest_optdesc,
	.non_option_args = 1,

	.run  = memtest_run,
};
