/*
 * cmd/reg.c : implements the JTAGER reg command.
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: cmd_reg.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "jtager.h"
#include "target.h"

#define REG_SUBCMD_LIST 	0x01
#define REG_SUBCMD_READ_ALL 	0x02
#define REG_SUBCMD_READ 	0x04
#define REG_SUBCMD_WRITE 	0x08

static int reg_subcmd = 0;

static int list_flag = 0;      /* --list option */
static char *name_str = NULL;  /* --name option */
static char *value_str = NULL; /* --value option */

static struct option reg_options[] = {
	{"list", no_argument, &list_flag, 1},
	{"name", required_argument, 0, 'n'},
	{"value", required_argument, 0, 'v'},
	{0, 0, 0, 0}
};
static const char *reg_short_options = "nv";

static struct option_desc reg_optdesc[] = {
	{"list", "Print a list of all available core register."},
	{"name", "The name of the target core registers, such as \"r10\"."},
	{"value", "The new value to be written to the target core register."},
	{NULL, NULL}
};

static int do_reg_list(void)
{
	int i;
	core_register_t *reg = NULL;

	reg = target->regs;
	i = 1;
	if (reg == NULL) {
		printf("Error: There are no available core "\
			"registers for you target.\n");
		return -1;
	}

	printf("The name list of all available core registers "
		"for your target:\n");
	while(reg && reg->name) {
		printf("%s", reg->name);
		if ((i % 8) == 0)
			printf("\n");
		else
			printf("\t");
		i++;
		reg++;
	} /* end of while(...) */
	printf("\n");
	return 0;
} /* end of do_reg_list() */

static int do_reg_read_all()
{
	int ret, i = 1;
	core_register_t *reg = NULL;

	ret = t_op->get_core_state();
	if (ret) {
		printf("Error:%s", err_str(ret));
		return ret;
	}
	
	/* print all registers */
	reg = target->regs;
	while (reg && reg->name) {
		printf("%s = 0x%.8lX", reg->name, reg->value);
		if ((i % 3) == 0) {
			printf("\n");
		} else if (strlen(reg->name) <= 2) { 
			printf("\t\t");
		} else
			printf("\t");
		reg++;
		i++;
	}
	printf("\n");
	return 0;
} /* end of do_reg_read_all(...) */

static int do_reg_read(core_register_t *reg)
{
	int ret;
	
	ret = t_op->register_read(reg);
	if (ret) {
		printf("Error:%s", err_str(ret));
	} else {
		printf("%s = 0x%.8lX\n", reg->name, reg->value);
	}

	return ret;
} /* end of do_reg_read() */

static int do_reg_write(core_register_t *reg, u32 newval)
{
	int ret;

	ret = t_op->register_write(reg, newval);
	if (ret) {
		printf("Error:%s", err_str(ret));
	} else {
		printf("%s register has been set new value 0x%.8lX, "\
			"successfully.\n", reg->name, newval);
	}

	return ret;
} /* end of do_reg_write() */

static int reg_run(int argc, char **argv)
{
	int option_index = 0, ret = 0;
	u32 newval = 0;
	core_register_t *reg = NULL;
	char *invalid_arg = NULL;

	/*
	 * Step 1 - reset all option flag variables and option value strings
	 */
	list_flag = 0;
	name_str = value_str = NULL;
	reg_subcmd = 0;
	
	/*
	 * Step 2: decode command options and non-option arguments
	 */
	while (1) {
		ret = getopt_long(argc, argv, reg_short_options,
				reg_options, &option_index);

		/* detec the end of the options */
		if (ret == -1) {
			ret = 0;
			break;
		}
		
		switch (ret) {
		case 0:
			/* this option has set a flag, do nothing else now */
			if (reg_options[option_index].flag != 0)
				break;
			else
				goto syntax_err;
		case 'n':
			/* --name option */
			if (name_str)
				goto syntax_err;
			name_str = optarg;
			break;
		case 'v':
			/* --value option */
			if (value_str)
				goto syntax_err;
			value_str = optarg;
			break;
		case '?':
			/* Fall through to default ...
			 * But before we print our syntax error message,
			 * getopt_long will print an error message, first.
			 */
		default:
			ret = -1;
			goto syntax_err;
		} /* end of switch (...) */
	} /* end of while (1) */
	
	/* Try to get non-option argument */
	if (optind < argc) {
		if (value_str)
			goto syntax_err;
		value_str = (char *)argv[argc-1];
	}

	/*
	 * Step 3 - set sub-command to be taken.
	 * NOTE: Only one sub-command can be executed.
	 */
	if (argc == 1)
		reg_subcmd = REG_SUBCMD_READ_ALL;
	
	if (list_flag) {
		if (reg_subcmd || argc != 2)
			goto syntax_err;
		reg_subcmd = REG_SUBCMD_LIST;
	} /* end of if (reg_list_flag) */

	/* read or write */
	if (name_str) {
		if (reg_subcmd)
			goto syntax_err;
		reg = reg_index(name_str);
		if (reg == NULL)
			goto unknown_reg;
		if (value_str == NULL) {
			reg_subcmd = REG_SUBCMD_READ;
		} else {
			/* translate new value string */
			invalid_arg = value_str;
			ret = str2int(value_str, &newval);
			if (ret)
				goto arg_err;
			reg_subcmd = REG_SUBCMD_WRITE;
		}
	} /* end of if (reg_name) */
	
	/*
	 * Step 4 - go
	 */
	switch (reg_subcmd) {
	case REG_SUBCMD_LIST:
		ret = do_reg_list();
		break;
	case REG_SUBCMD_READ_ALL:
		ret = do_reg_read_all();
		break;
	case REG_SUBCMD_READ:
		ret = do_reg_read(reg);
		break;
	case REG_SUBCMD_WRITE:
		ret = do_reg_write(reg, newval);
		break;
	default:
		goto syntax_err;
	} /* end of switch (reg_subcmd) */

	return ret;
syntax_err:
	return print_syntax_error(argv[0]);
arg_err:
	return print_arg_error(argv[0], invalid_arg);
unknown_reg:
	printf("%s: Unknown core register name: %s\n", argv[0], name_str);
	printf("Please use the command \"reg --list\" to get a list of "\
		"all available core registers name for your target.\n");
	return -1;
} /* end of reg_run() */

cmd_t reg_cmd = {
	.name = "reg",
	.desc = "Read/write the core registers of the target in the debug state",
	.alias = NULL,
	.usage = "reg [--list]\n\treg --name=regname [[--value=]newvalue]",
	.verbose = "While the target is the debug state, this command can "\
		"print the\n\tvalues of all core registers, or the value of "\
		"the specified\n\tregister, or set a new value for the " \
		"specified register of target.",
	.long_options = reg_options,
	.optdesc = reg_optdesc,
	.non_option_args = 1,
	
	.run = reg_run
};
