/*
 * cmdline/command.c : implements the JTAGER commands parser.
 * 
 * Copyright (C) 2003-2004, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: command.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

#include <unistd.h>
#define _GNU_SOURCE /* for getline() function */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>
#include <readline/readline.h>

#include "jtager.h"

/* basic commands group */
static cmd_t basic_group = {
	"group", "BASIC COMMANDS GROUP", NULL, 0, NULL, NULL
};
extern cmd_t help_cmd;
extern cmd_t reset_cmd;
extern cmd_t idcode_cmd;
extern cmd_t exit_cmd;

/* debug commands group */
static cmd_t debug_group = {
	"group", "DEBUG COMMANDS GROUP", NULL, 0, NULL, NULL
};
extern cmd_t ice_cmd;
extern cmd_t halt_cmd;
extern cmd_t restart_cmd;
extern cmd_t execbat_cmd;
extern cmd_t reg_cmd;
extern cmd_t memcmp_cmd;
extern cmd_t memcpy_cmd;
extern cmd_t memcsum_cmd;
extern cmd_t memdump_cmd;
extern cmd_t memset_cmd;
extern cmd_t memtest_cmd;
extern cmd_t flprobe_cmd;
extern cmd_t flerase_cmd;
extern cmd_t flwrite_cmd;

/* all JTAGER commands list */
cmd_t *cmds_tbl[] = {
	&basic_group, /* group separator */
	&help_cmd,
	&reset_cmd,
	&idcode_cmd,
	&exit_cmd,

	&debug_group, /* group separator */
	&ice_cmd,
	&halt_cmd,
	&restart_cmd,
	&execbat_cmd,
	&reg_cmd,

	&memcmp_cmd,
	&memcpy_cmd,
	&memcsum_cmd,
	&memdump_cmd,
	&memset_cmd,
	&memtest_cmd,

	&flprobe_cmd,
	&flerase_cmd,
	&flwrite_cmd,

	NULL /* the last element of array must be NULL */
};

cmd_t * command_index(const char *cmdname)
{
	int i = 0;
	cmd_t *cmd;
	
	while (1) {
		cmd = cmds_tbl[i++];
		if (cmd == NULL)
			break;  /* This is the last array element */
		if ((strcasecmp(cmdname, cmd->name) == 0) ||
			((cmd->alias) && (strcasecmp(cmdname, cmd->alias) == 0)))
			break;
	}
	
	return cmd;
} /* end of cmd_index(...) */

/*
 * match_option_args - match all the options' name of the command line
 * @non_option_args: the number of non-option arguments permitted in
 * 					the cmd_argv[] string array.
 * 
 * NOTE: The reason that we need this function is that there is a bug in
 * glibc function - getopt_long().
 */
int match_option_args(int cmd_argc,
			char *cmd_argv[],
			struct option *opts,
			int non_option_args)
{
	int i, j;
	int found, retval;
	char *optname1, *optname2;
	struct option *opt = NULL;
	int nonoption_nr = 0;
	
	/* do not scan argv[0], it is command name */
	found = 1;
	retval = 0;
	for (i = 1; i < cmd_argc; i++) {
		if (!cmd_argv[i])
			break;

		/* 
		 * If non_option_args = 1, non-option arguments are permitted,
		 * therefore we don't care about them.
		 * If non_option_args = 0, non-option arguments are forbidden,
		 * therefore we return error.
		 */
		if (cmd_argv[i][0] != '-') {
			nonoption_nr++;
			if (nonoption_nr > non_option_args) {
				found = 0;
				break;
			}
			continue;
		}

		/*
		 * OK, now cmd_argv[i] is sure to be an option argument.
		 * Let's match it ...
		 */
		optname1 = strdup(cmd_argv[i]);
		if (!optname1) {
			retval = -1;
			break;
		}

		/*
		 * Get rid of other chars except option name ...
		 * NOTE: in the cmd_argv[i] string, the sub-string after
		 * the first char '=' is its argument value.
		 */
		optname2 = optname1 + 1; /* strip off the leading '-' char */
		if (*optname2 == '-')
			optname2++;
		j = 0;
		while (optname2[j] != '\0') {
			if (optname2[j] == '=') {
				optname2[j] = '\0';
				break;
			}
			j++;
		}
		
		/* option name match */
		found = 0;
		j = 0;
		while (1) {
			opt = &opts[j];
			if (opt->name == NULL)
				break;
			if (strcasecmp(opt->name, optname2) == 0) {
				found = 1;
				break;
			}
			j++;
		}
		
		free(optname1);
		if (found == 0)
			break;
	} /* end of for (i = 0...) */
	
	if (found == 0)
		retval = i;
	return retval;
} /* end of match_option_args(...) */

#define MAX_OPTIONS_NUM 	10
static int cmdline_argc = 0;
static char *cmdline_argv[MAX_OPTIONS_NUM]; /* the last element must be NULL */

/*
 * parse_cmdline - the commands dispatcher of the JTAGER commands parser.
 */
int parse_cmdline(char * cmdline)
{
	char *param, *cmdname;
	cmd_t * cmd;
	int i, retval;
	
	if (!cmdline)
		return -1;

	/* translate all characters to lower chars */
	i = 0;
	while (i < strlen(cmdline)) {
		cmdline[i] = tolower(cmdline[i]);
		i++;
	}

	/* get the command name */
	cmdname = strtok(cmdline, " \n");
	cmdline_argv[0] = cmdname; /* argv[0] is always the command name */
	cmdline_argc = 1;

	/* get others command options */
	while (1) {
		param = strtok(NULL, " \n");
		cmdline_argv[cmdline_argc] = param;
		if (!param)
			break;
		cmdline_argc++;
	} /* end of while (1) */

	/* index the specified command */
	cmd = cmd_index(cmdname);
	if (cmd == NULL) {
		printf("%s: Unknown command.\n", cmdname);
		return -1;
	}

	/*
	 * check option name ...
	 * Damn glibc bug.
	 */
	if (cmd->long_options) {
		retval = match_option_args(cmdline_argc,
					cmdline_argv,
					cmd->long_options,
					cmd->non_option_args);
		if (retval == -1) {
			perror(NULL);
			return -1;
		} else if (retval) {
			printf("%s: unrecognized option: \"%s\".\n",
				cmdname, cmdline_argv[retval]);
			return -1;
		}
	}

	/*
	 * before going, we must set optind to zero, which will let
	 * getopt_long() function initialize itself.
	 */
	optind = 0;
	retval = -1;
	if (cmd->run)
		retval = cmd->run(cmdline_argc, cmdline_argv);
	else
		printf("%s: Un-implemented command.\n", cmdname); 
	
	return retval;
} /* end of parse_cmdline(...) */

/*
 * Command line error handler
 */
int print_syntax_error(char *cmdname)
{
	printf("%s: syntax error.\n", cmdname);
	return -1;
}

int print_arg_error(char *cmdname, char *invalid_arg)
{
	printf("%s: invalid argument value: %s\n", cmdname, invalid_arg);
	return -1;
}

/**
 * get_memsize - get memory block size
 * @size_str: the memory size string
 * 
 * NOTE: 100b/100B - 100 bytes.
 *       100s/100S - 100 half-words, or 200 bytes
 *       100w/100W - 100 words, or 200 half-words, or 400 bytes
 *
 * Return value: If success, return BYTE_FMT, HALFWORD_FMT or WORD_FMT
 *               If failure, return -1 : unkown size format, or bad size string
 */
int get_memsize(char *size_str, u32 *size)
{
	int len, sizefmt;
	char ch;

	if (size_str == NULL) {
		if (size)
			*size = 1;
		return WORD_FMT; /* default: 32-bit word */
	}
	
	/* The final character of the size string is the indicator */
	len = strlen(size_str);
	ch = tolower(size_str[len-1]);
	if (ch == BYTE_INDICATOR) {
		size_str[len-1] = '\0';
		sizefmt = BYTE_FMT;
	} else if (ch == HALFWORD_INDICATOR) {
		size_str[len-1] = '\0';
		sizefmt = HALFWORD_FMT;
	} else if (ch == WORD_INDICATOR) {
		size_str[len-1] = '\0';
		sizefmt = WORD_FMT;
	} else {
		sizefmt = WORD_FMT; /* default: 32-bit word */
	}
	
	if (size) {
		if (str2int(size_str, size))
			return -1; /* bad size string */
	}

	return sizefmt;
}

/**
 * ask_yesno - ask users 'yes' or 'no'
 *
 * Return value: 1 - yes, 0 - no.
 */
int ask_yesno(const char *prompt)
{
	int ret, len;
	char *answer;

repeat:
	printf("%s (ENTER=y) (y/n): ", prompt);
	answer = NULL;
	len = 0;
	len = getline(&answer, &len, stdin);
	
	if ((len == 1 && (ret = strcasecmp(answer, "\n")) == 0)
		|| (len == 2 && (ret = strcasecmp(answer, "y\n")) == 0)
		|| (len == 4 && (ret = strcasecmp(answer, "yes\n")) == 0))
		ret = 1;
	else if ((len == 2 && (ret = strcasecmp(answer, "n\n")) == 0)
		|| (len == 4 && (ret = strcasecmp(answer, "no\n")) == 0))
		ret = 0;
	else {
		free(answer);
		goto repeat;
	}
	
	if (answer)
		free(answer);
	return ret;
} /* end of ask_yesno() */
