/*
 * jtag/jtag.c : Implement JTAGER's jtag control interface
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>, All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: jtag.c,v 1.2 2004/10/17 08:04:28 zhanrk Exp $ */

/*
 * References:
 * [1] IEEE Std 1149.1-1990, <<IEEE Standard Test Access Port and 
 *      Boundary-Scan Architecture>>
 * [2] IEEE Std 1149.1b-1994, <<Supplement to IEEE Std 1149.1-1990>>
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>	/* for direct read/write I/O port */

#include "jtager.h"
#include "jtag.h"
#include "target.h"
#include "types.h"

/*
 * The port I/O address of the Parallel port (LPTn), which we are using to
 * connect the JTAG dongle. By defaut, it is LPT1 whose I/O address is from
 * 0x378 to 0x380.
 */
unsigned int par_dataport = 0x378; 	/* data port */
unsigned int par_statusport = 0x379; /* status port */

/* The local copy of LPTn data port */
static unsigned char dataport_value = 0; /* JTAG_TRST_MASK */

/* The TAP controller is in Run-Test/Idle state ? */
static unsigned char jtag_idle = 0;

/*
 * set_tms - set new value for the TMS signal
 * @tms: the new logic value for TMS signal, it must be 0 or 1.
 *
 * NOTE: The variable "dataport_value" will remember your new setting
 * to the value of TMS signal.
 */
static void set_tms(int tms)
{
	if (tms)
		dataport_value |= JTAG_TMS_MASK;
	else
		dataport_value &= ~JTAG_TMS_MASK;
	outb(dataport_value, par_dataport);
} /* end of set_tms(...) */

/*
 * pulse_tck - Pulse the TCK signal
 *
 * NOTE: This function will ouput a 1 on TCK signal. But we don't modify
 * the value of the variable "dataport_value".
 */
static void pulse_tck(void)
{
	outb(dataport_value | JTAG_TCK_MASK, par_dataport);
} /* end of pulse_tck() */

/*
 * set_tdi - Set new value for the TDI pin of the jtag interface
 * @tdi_val: the new logic value for TDI signal, it must be 0 or 1.
 */
static void set_tdi(int tdi_val)
{
	if (tdi_val)
		dataport_value |= JTAG_TDI_MASK;
	else
		dataport_value &= ~JTAG_TDI_MASK;
	outb(dataport_value, par_dataport);
} /* end of set_tdi(...) */

/*
 * read_tdo - Read the logic value of TDO signal
 * 
 * NOTE: Because the pin11 of LPT is inverted by hardware, therefore, if we
 * read a 1 from pin11 of LPT, it means the logic value of TDO signal is 0.
 * By contrast, if we read a 0 from pin11 of LPT, it means the logic value
 * of TDO signal is 1.
 */
static int read_tdo(void)
{
	if (inb(par_statusport) & JTAG_TDO_MASK)
		return 0;
	else
		return 1;
} /* end of get_tdo() */

/*
 * enter_next_state - make the TAP controller enter the next adjacent state
 * @tms: the new logic value for TMS signal, it must be 0 or 1.
 * 
 * NOTE: Please see <<IEEE Std 1149.1-1990>> for the details of the TAP
 * controller's state machine.
 */
void enter_next_state(int tms)
{
	set_tms(tms);
	pulse_tck();
} /* end of enter_next_state(...) */

/*
 * jtag_access_reg - Read/write the jtag test instruction register,
 *		     or  the jtag test data register.
 * @bitnr    : the bit number to be written or read.
 * @write_buf: the buffer containing the data to be written into jtag register.
 * @read_buf : when the function returns, the buffer will contain
 * 	       the data read out from jtag register.
 * 
 * NOTE: We assume that the caller makes the jtag TAP controller in the
 * Shift-DR/Shift-IR state. When this function returns, it makes the jtag TAP
 * controller entering Exit1-DR/Exit1-IR state. 
 * 
 * It is the duty of the caller to make the TAP controller enter into the 
 * Update-DR/Update-IR state, by hand.
 *
 * When the TAP controller is in the Shift-DR state on entry, we will write
 * data into the data register, or read data from the data register.
 * Likewise, when in Shift-IR state on entry, we will write data into the
 * instruction register.
 */ 
static int jtag_access_reg(int bitnr, const u32 *write_buf, u32 *read_buf)
{
	/*
	 * bitcnt    : the counter for the bit that we are reading/writing.
	 * wordcnt : the counter for the word that we are reading/writing.
	 * wordnr   : the word (32 bits) number that we want to read/write.
	 *               It is calculated by the paramter "bitnr".
	 */
	int bitcnt, wordcnt, wordnr;
	u32 writedata, readdata, value;

	if (bitnr == 0 || (!write_buf && !read_buf))
		return -ERR_INVALID_PARAM;

	/* Calculate the word number to read/write */
	wordnr = bitnr/32;
	if ((bitnr % 32) != 0)
		wordnr ++;
	
	writedata = readdata = wordcnt = 0;
	for (bitcnt = 0; bitcnt < bitnr; bitcnt++) {
		/*
		 * Shift in one bit into TDI signal......
		 */

		/* Every 32 bits, we fetch a word from the write buffer */
		if ((bitcnt & 0x1f) == 0) {
			wordcnt = bitcnt/32; /* calculate the word counter */
			if (write_buf)
				writedata = write_buf[wordcnt];
			else
				writedata = 0;
		}
		
		value = 0; /* JTAG_TRST_MASK; */
		if (writedata & 0x01)
			value |= JTAG_TDI_MASK;

		/* If last bit, put TMS=1 to the TAP controller
		 * enter the Exit1-DR/Exit1-IR state.
		 */
		if (bitcnt == (bitnr-1))
			value |= JTAG_TMS_MASK;

		outb(value, par_dataport);
		outb(value | JTAG_TCK_MASK, par_dataport);

		writedata >>= 1;
		
		/*
		 * Shift out one bit from TDO signal......
		 * NOTE: the pin11 of LPT is inverted by hardware!
		 */
		if (read_tdo())
			readdata |= (1 << (bitcnt & 0x1f));

		/* 
		 * If having read a word, store it back. 
		 * Or if it is the last bit, store readdata back.
		 */
		if (((bitcnt & 0x1f) == 0x1f) || (bitcnt == (bitnr - 1))){
			if (read_buf)
				read_buf[wordcnt] = readdata;
			readdata = 0;
		} 
	} /* end of for (bitcnt...) */

	return 0;
} /* end of jtag_access_reg(...) */

/*
 * jtag_rw_dreg - read/write jtag test data register from TDO signal
 * @bitnr: the bit number to read
 * @write_buf: the buffer containing the data to be written in
 * @read_buf: the buffer containing the data read out.
 * 
 * NOTE: We assume that the caller makes the jtag TAP controller in the
 * Run-Test/Idle or Update-DR/Update-IR state. When this function returns,
 * it makes the jtag TAP controller staying in the Update-DR state.
 */
int jtag_rw_dreg(int bitnr, const u32 *write_buf, u32 *read_buf)
{
	int retval;
	
	if (bitnr <= 0)
		return -ERR_INVALID_PARAM;
	
	/*
	 * Whenever the TAP controller is in the Run-Test/Idle state, or in the
	 * Update-DR/Update-IR state, putting TMS = 1 wil make TAP enter
	 * the Select-DR-Scan state.
	 */
	enter_next_state(1); /* Goto Select-DR-Scan */
	jtag_idle = 0;
	
	enter_next_state(0); /* Select-DR-Scan --> Capture-DR */
	enter_next_state(0); /* Capture-DR --> Shift-DR */

	retval = jtag_access_reg(bitnr, write_buf, read_buf);

	enter_next_state(1); /* Exit1-DR --> Update-DR */

	/* we always go via Idle/Run for data register stuff */
	enter_next_state(0); /* Update-DR --> Run-Test/Idle */
	jtag_idle = 1;
	return retval;
}

/*
 * jtag_write_ireg - write jtag test intructions into the intruction register
 * @jtag_instruction: the jtag public instruction to be written.
 *
 * NOTE: We assume that the caller makes the jtag TAP controller in the
 * Run-Test/Idle or Update-DR/Update-IR state.
 * 
 * When this function returns, it makes the jtag TAP controller entering into
 * the Update-IR state. But if the jtag instruction written in is RESTART
 * instruction, then the TAP controller will return back to the Run-Test/Idle
 * state on exit.
 * 
 * During the TAP controller is in the Capture-IR state, the fixed value b0001
 * is loaded into the instruction register. Therefore, we always read out the
 * fixed value b0001 from TDO during the Shift-IR state. If not, something is
 * sure to be wrong. 
 */
int jtag_write_ireg(u32 jtag_instruction)
{
	int bitnr = target->instruction.bitnr;
	int retval;
	u32 fixed;
	
	/*
	 * Whenever the TAP controller is in the Run-Test/Idle state, or in the
	 * Update-DR/Update-IR state, putting TMS = 1 wil make TAP enter
	 * the Select-DR-Scan state.
	 */
	enter_next_state(1); /* Goto Select-DR-Scan */
	jtag_idle = 0;

	enter_next_state(1); /* Select-DR-Scan --> Select-IR-Scan */
	enter_next_state(0); /* Select-IR-Scan --> Capture-IR */
	enter_next_state(0); /* Capture-IR --> Shift-IR */
	
	retval = jtag_access_reg(bitnr, &jtag_instruction, &fixed);
	enter_next_state(1); /* Exit1-IR --> Update-IR */
	if (retval)
		return retval;

	if (jtag_instruction == JTAG_RESTART) {
		enter_next_state(0); /* Update-IR --> Run-Test/Idle */
		jtag_idle = 1;
	}
	
	/* It should be some fixed known value, which is specific to
	 * target type. For ARM7TDMI target core, it is b0001.
	 */	
	if (fixed != target->instruction.readout)
		retval = -ERR_JTAG_INSTRUCTION_FAILED;
	return retval;
} /* end of jtag_write_ireg(...) */

/*
 * jtag_idcode - Read the Device Identification Code
 * 
 * NOTE: We assume that the TAP controller is in the Run-Test/Idle state
 * or Update-DR/Update-IR state before this function is called.
 */
int jtag_idcode(u32 *idcode)
{
	int retval;
	int bitnr = target->idcode.bitnr; /* 32 bits */
	
	/*
	 * Write the JTAG instruction IDCODE into the instruction register.
	 */
	retval = jtag_write_ireg(JTAG_IDCODE);
	if (retval)
		return retval;

	/*
	 * Read out the Device Identification Code from the data register.
	 * The length of Device Id is 32 bits.
	 */
	retval = jtag_rw_dreg(bitnr, NULL, idcode);
	if (retval)
		return retval;

	/* restart */
	retval = jtag_write_ireg(JTAG_RESTART);
	if (retval)
		return retval;

	/*
	 * The bit 0 of IDCODE must be 1, according to the JEDEC Standard.
	 * If not, it is an invalid IDCODE value.
	 */
	if ((*idcode & 0x01) == 0) {
		retval = -ERR_JTAG_INVALID_IDCODE;
		*idcode = 0;
	} else {
		target->idcode.readout = *idcode; /* save it */
		retval = 0;
	}
	
	return retval;
} /* end of jtag_idcode() */

/*
 * jtag_select_scanchain - select the active scan chain between TDI and TDO
 *
 * NOTE: By default, scan chain 0 is the default one.
 * 
 * For ARM7TDMI core, the scan path select register is 4-bits long.
 * And SCAN_N instruction always outputs the fixed value - b1000.
 *
 * For ARM9TDMI core, the scan path select register is 5-bits long.
 * And SCAN_N instruction always outputs the fixed value - b10000.
 */
int jtag_select_scanchain(u32 chain_num)
{
	u32 fixed = 0;
	int retval;
	/* the bit number of scan path select register */
	int bitnr = target->scanpath.bitnr;
	
	if (target->active_sc == chain_num)
		return 0;
	if (chain_num > (target->sc_num - 1))
		return -ERR_JTAG_NONEXISTED_SCANPATH;
	
	retval = jtag_write_ireg(JTAG_SCAN_N);
	if (retval)
		return retval;
	retval = jtag_rw_dreg(bitnr, &chain_num, &fixed);
	if (retval)
		return retval;
	
	/* It should be some fixed known value, which is specific to
	 * target type. For ARM7TDMI target core, it is b1000.
	 */
	if (fixed != target->scanpath.readout)
		retval = -ERR_JTAG_SCANPATH_FAILED;
	else
		target->active_sc = chain_num;
	
	return retval;
} /* end of jtag_select_scanchain() */

int jtag_bypass(void)
{
	u32 fixed = 1; /* any non-zero value */
	int retval;

	retval = jtag_write_ireg(JTAG_BYPASS);
	if (retval)
		return retval;
	retval = jtag_rw_dreg(target->bypass.bitnr, NULL, &fixed);
	if (retval)
		return retval;
	
	if (fixed != target->bypass.readout)
		retval = -1;
	return retval;
} /* end of jtag_bypass() */

/* 
 * jtag_reset - reset the JTAG TAP controller
 *
 * NOTE: After reset, the JTAG TAP controller will enter Run-Test/Idle state.
 */
int jtag_reset(void)
{
	int count;
	
	/* You MUST be root to use read/write parallel port */
	printf("Using parallel data port and status port -- "\
		"0x%X and 0x%X\n", par_dataport, par_statusport);
	if (ioperm (par_dataport, 3, 1)) {
		perror(PACKAGE);
		printf("Error: Hoho, You have no previlege to read/write "\
			"the parallel ports\n");
		return -1;
	}

	outb(JTAG_TRST_MASK, par_dataport); /* LOW */
	usleep(100);
	outb(0, par_dataport); /* HIGH */
	usleep(100);

	/* 
	 * No matter what the original state of the TAP controller, it will
	 * enter the Test-Logic-Reset state, when TMS is held high for at
	 * least five rising edges of TCK. The controller remains in this
	 * state while TMS is high.
	 */
	for (count = 0; count <= 4; count++) {
		set_tms(1);
		pulse_tck();
	}

	/* Set TMS to 0, make TAP controller into Run-Test/Idle state.
	 * Once entering Run-Test/Idle state, it will remain in this state
	 * as long as TMS is held low.
	 */
	set_tms(0);
	pulse_tck(); /* Test-Logic-Reset ==> Run-Test/Idle */
	set_tms(0);
	pulse_tck(); /* Stay here: Run-Test/Idle ==> Run-Test/Idle */

	jtag_idle = 1;
	printf("Jtag interface has been reset successfully!\n");
	return 0; /* success */
} /* end of jtag_reset() */
