/*
 * main.c : the main source file for JTAGER
 *
 * Copyright (C) 2003, Rongkai Zhan <zhanrk@163.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: main.c,v 1.3 2004/10/17 13:54:32 zhanrk Exp $ */

/*
 * References:
 * [1] IEEE Std 1149.1-1990, <<IEEE Standard Test Access Port and 
 *      Boundary-Scan Architecture>>
 * [2] IEEE Std 1149.1b-1994, <<Supplement to IEEE Std 1149.1-1990>>
 * [3] <<ARM Architecture Reference Manual>>
 * [4] <<ARM7TDMI Technical Reference Manual>> Rev 4
 * [5] <<ARM9TDMI Technical Reference Manual>> Rev 3
 * [6] Application Note 28, <<The ARM7TDMI Debug Architecture>>
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "jtager.h"
#include "jtag.h"
#include "target.h"
#include "flash.h"

/*
 * The character buffer for jtager command line input...
 * It is terminated by NULL character '\0', therefore we get 2
 */
static const char *prompt = "JTAGER:> ";
static const char *etc_dir = "/etc/jtager";
const char *jtager_history = "/etc/jtager/history";

static char *err_desc[] = {
	"\tunknown error.\n", /* ERR_COMMON */

	"\tInvalid function parameters.\n", /* ERR_INVALID_PARAM */

	/* ERR_JTAG_INSTRUCTION_FAILED */
	"\tFailed to read/write the JTAG instruction register. The JTAG"\
	"\n\tinstruction register always outputs a fixed known value during"\
	"\n\tthe CAPTURE-DR stage. E.g. for ARM7TDMI target, its JTAG"\
	"\n\tinstruction register always outputs b0001. But now we read a"
	"\n\tdifferent value. Something is wrong.\n",

	/* ERR_JTAG_INVALID_IDCODE */
	"\tWe read an invalid idcode value from the JTAG IDCODE register."\
	"\n\tThe bit[0] of idcode value must be 1.\n",

	/* ERR_JTAG_NONEXISTED_SCANPATH */
	"\tYour specified scan chain is non-existed on your target.\n",

	/* ERR_JTAG_SCANPATH_FAILED */
	"\tFailed to select your specified scan chain. The JTAG scan path"\
	"\n\tselect register always outputs a fixed known value during the"\
	"\n\tCAPTURE-DR stage. E.g. for ARM7TDMI target, its JTAG scan path"\
	"\n\tselect register always outputs b1000. But now we read a"\
	"\n\tdifferent value. Something is wrong.\n",

	/* ERR_TARGET_NONEXISTED_ICE_REG */
	"\tYour specified ICE register is non-existed on your target. Maybe"\
	"\n\tyou have specified an illegal ICE register address.\n",

	/* ERR_TARGET_HALT_FAILED */
	"\tFailed to halt the target and make it into the debug state.\n",

	/* ERR_TARGET_IS_RUNNING */
	"\tThe target is running. Please use \"halt\" command to halt it.\n",

	/* ERR_TARGET_IN_THUMB_MODE */
	"\tThe target is in THUMB mode, which is not implemented by JTAGER.\n",

	/* ERR_TARGET_UNKNOWN_REG */
	"\tUnknown core register name.\n",

	/* ERR_TARGET_UNKNOWN_ICE_REG */
	"\tUnknown ICE register name.\n",

	/* ERR_TARGET_INVALID_CPSR */
	"The CPSR register's value is undetermined and invalid. Before we"\
	"\n\tcan read or write SPSR register, we must have have an valid CPSR"\
	"\n\tregister value. And you can use command \"reg cpsr\" to get it.\n",

	/* ERR_TARGET_NO_SPSR */
	"\tThe target is in USER mode, so it has no SPSR register.\n",

	/* ERR_FLASH_TIMEOUT */
	"\tTime out.\n",

	NULL /* the last one must be NULL, to terminate the array */
};

char *err_str(int retval)
{
	int index = -retval - 1;
	
	return err_desc[index];
}

int main(void)
{
	char * cmdline = NULL;
	char * temp = NULL;
	
	/* Print our logo information */
	printf("*** Welcome to %s-%s ......\n", PACKAGE, VERSION);
	printf("*** Copyright (C) 2003-2004, "AUTHOR" <zhanrk@163.com>\n\n");
	printf(PACKAGE" is a free software, covered by the GNU General Public "\
		"License, and you\nare welcome to change it and/or distribute"\
		"copies of it under certain conditions.\n"\
		"There is absolutely no warranty for "PACKAGE".\n\n");
	printf("You can type 'help' command to get a list of all commands. "\
		"You aslo can\ntype 'help <command>' command to get a "\
		"verbose information about the command.\n"\
		"All commands are case-insensitive, and has GNU-style long "\
		"command options.\n\n");

	/* Initialize jtag */
	if (jtag_reset()) {
		printf("Error: %s: failure to initialize jtag ...\n", PACKAGE);
		exit(-1);
	}

	if (setup_config())
		exit(-1);

	/* Create the directory if it doesn't exists. */
	mkdir(etc_dir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	
	/* Initialize history and read history file */
	using_history();
	read_history(jtager_history);
	
	/* Main loop : parse user's command line */
	while(1) {
		if (cmdline) {
			free(cmdline);
			cmdline = NULL;
		}
		cmdline = readline(prompt);
		if (cmdline == NULL)
			continue;
		
		/* copy command line */
		temp = strdup(cmdline);
		if (!temp) {
			printf("Error: not enough memory.\n");
			continue;
		}
			
		/* strip the leading space */
		while(*temp == ' ')
			temp++;

		/*
		 * Those command lines, whose leading character is
		 * '\0' or '\n' or '\r', are treated as invalid command lines,
		 * and ignored directly.
		 */
		if ((temp[0] == '\0') || (temp [0] == '\n')
					|| (temp[0] =='\r')) {
			free(temp);
			continue;
		}

		/* update command history */
		add_history(temp);

		/* command line parsing... */
		parse_cmdline(temp);
	} /* end of while(1) */

	write_history(jtager_history);
	exit(0);
} /* end of main() */
